//  Program: CS2_UTL
//   Author: Carlo Kok
//  Purpose: Utility functions to support script component
//-------------------------------------------------------------------
unit IFS_UTL;
{$I IFS_DEF.INC}

interface


type

   // The following is the list of supported tokens.  If you add to
   // this list, you need to write all of the code to support the new
   // token.

  TIfPasToken = (
  {Errors}
    CSTI_SyntaxError,
    CSTI_CommentEOFError,
    CSTI_CharError,
    CSTI_StringError,
  {Tokens}
    CSTI_EOF,
    CSTI_Whitespace,
    CSTI_Comment,
    CSTI_Identifier,
    CSTI_SemiColon,
    CSTI_Comma,
    CSTI_Period,
    CSTI_Colon,
    CSTI_OpenRound,
    CSTI_CloseRound,
    CSTI_OpenBlock,
    CSTI_CloseBlock,
    CSTI_Assignment,
    CSTI_Equal,
    CSTI_NotEqual,
    CSTI_Greater,
    CSTI_GreaterEqual,
    CSTI_Less,
    CSTI_LessEqual,
    CSTI_Plus,
    CSTI_Minus,
    CSTI_Divide,
    CSTI_Multiply,
    CSTI_Integer,
    CSTI_Real,
    CSTI_String,
    CSTI_Char,
    CSTI_HexInt,
  {Identifiers}
    CSTII_and,
    CSTII_array,
    CSTII_begin,
    CSTII_case,
    CSTII_const,
    CSTII_div,
    CSTII_do,
    CSTII_downto,
    CSTII_else,
    CSTII_end,
    CSTII_for,
    CSTII_function,
    CSTII_if,
    CSTII_in,
    CSTII_mod,
    CSTII_not,
    CSTII_of,
    CSTII_or,
    CSTII_procedure,
    CSTII_program,
    CSTII_repeat,
    CSTII_set,
    CSTII_shl,
    CSTII_shr,
    CSTII_string,
    CSTII_then,
    CSTII_to,
    CSTII_type,
    CSTII_until,
    CSTII_uses,
    CSTII_var,
    CSTII_while,
    CSTII_with,
    CSTII_xor,
    CSTII_exit,
    CSTII_break
    );


  PIfPascalParser = ^TIfPascalParser;
  TIfPascalParser = record
    CurrTokenId: TIfPasToken;
    CurrTokenLen: LongInt;
    CurrTokenPos: LongInt;
    Text: PChar;
  end;
  PCs2PascalParser = ^TCs2PascalParser;
  TCs2PascalParser = TIfPascalParser;

procedure SkipWhiteSpace(data: PIfPascalParser);
{Go to the next nojunk token}

function GetToken(data: PIfPascalParser): string;
{Return the token}
procedure ParseToken(data: PIfPascalParser);
{Parse at active position}

type
  PIfListItem = ^TIfListItem;
  TIfListItem = record
    Next: PIfListItem;
    Prev: PIfListItem;
    Ptr: Pointer;
  end;
  TIfList = {$IFDEF CBUILDER}class{$ELSE}object{$ENDIF}
  Private
    FCount: LongInt;
    RootItem: PIfListItem;
    CurrItem: PIfListItem;
    currItemIdx: Longint;
  Public
    function Count: LongInt;
    function GetItem(Nr: LongInt): Pointer;
    procedure SetItem(Nr: LongInt; P: Pointer);
    procedure Add(P: Pointer);
    procedure Remove(P: Pointer);
    procedure Delete(Nr: Longint);
    procedure Clear;
    constructor Create;
    destructor Destroy; {$IFDEF CBUILDER}Override; {$ENDIF}
  end;

  TIfStringList = {$IFDEF CBUILDER}class{$ELSE}object{$ENDIF}
  Private
    List: TIfList;
  Public
    function Count: LongInt;
    function GetItem(Nr: LongInt): string;
    procedure SetItem(Nr: LongInt; const s: string);
    procedure Add(const P: string);
    procedure Delete(NR: LongInt);
    procedure Clear;
    constructor Create;
    destructor Destroy; {$IFDEF CBUILDER}Override; {$ENDIF}
  end;

function FastUpperCase(const s: string): string;     // Fast uppercase
function FastLowerCase(const s: string): string;     // Fast lowercase
function Fw(const S: string): string;                // First word
procedure Rs(var S: string);                         // Remove space left



implementation


//-------------------------------------------------------------------
function Fw(const S: string): string;          //  First word
var
  x : integer;
begin
  x := pos(' ',s);
  if x > 0
  then Fw := Copy(S, 1, x-1)
  else Fw := S;
end;
//-------------------------------------------------------------------
procedure Rs(var S: string);                  // Remove space
var
  x : integer;
begin               
  if s>'' then
  begin
     x := 1;
     while (copy(s,x,1)=' ') do inc(x);
     if x >1
     then s := copy(s,x,(length(s)-x)+1);
  end;
end;
//-------------------------------------------------------------------
constructor TIfList.Create;
begin
{$IFDEF CLASSLIST}
  inherited Create;
{$ENDIF}
  FCount := 0;
  RootItem := nil;
  CurrItem := nil;
  curritemidx := -1;
end;
//-------------------------------------------------------------------
procedure TIfList.Add(P: Pointer);
var
  w: PIfListItem;
begin
  if RootItem = nil then
  begin
    new(rootitem);
    rootitem^.prev  := nil;
    rootitem^.next  := nil;
    rootitem^.Ptr   := p;
    CurrItem        := rootitem;
    curritemidx     := 0;
    inc(fcount);
  end else
  begin
    while assigned(CurrItem^.next) do
    begin
      inc(CurrItemIdx);
      curritem := curritem^.next;
    end;
    new(w);
    curritem^.next := w;
    w^.Next        := nil;
    w^.prev        := curritem;
    w^.Ptr         := p;
    inc(fcount);
  end;
end;
//-------------------------------------------------------------------
procedure TIfList.Delete(Nr: Longint);
begin
  if not Assigned(RootItem) or (nr >= fCount) or (nr < 0)then
    Exit;
  if nr < curritemidx then
  begin
    CurrItem    := RootItem;
    CurrItemIdx := 0;
  end;
  while curritemidx < nr do
  begin
    CurrItem := CurrItem^.Next;
    Inc(CurrItemIdx);
  end;
  if assigned(curritem^.prev) then
  begin
    curritem^.prev^.next := curritem^.next;
    if assigned(CurrItem^.next) then
      curritem^.next^.prev := currItem^.prev;
  end
  else

  begin
    rootitem := curritem^.next;
    if assigned(curritem^.next) then
      curritem^.next^.prev := nil;
  end;
  dispose(curritem);
  currItemIdx := 0;
  CurrItem    := rootitem;
  dec(fcount);
end;
//-------------------------------------------------------------------
procedure TIfList.Remove(P: Pointer);
begin
  CurrItem := rootitem;
  while assigned(curritem) do
  begin
    if curritem^.Ptr = p then
    begin
      if assigned(curritem^.prev) then
      begin
        curritem^.prev^.next := curritem^.next;
        if assigned(CurrItem^.next) then
          curritem^.next^.prev := currItem^.prev;
      end
      else

      begin
        rootitem := curritem^.next;
        if assigned(curritem^.next) then
          curritem^.next^.prev := nil;
      end;
      dispose(curritem);
      dec(fcount);
      break;
    end;
    curritem := curritem^.next;
  end;
  currItemIdx := 0;
  CurrItem := rootitem;
end;
//-------------------------------------------------------------------
procedure TIfList.Clear;
begin
  curritem := rootitem;
  while assigned(curritem) do
  begin
    if assigned(curritem^.next) then
       begin
         curritem := curritem^.next;
         dispose(curritem^.prev);
       end
    else
       begin
          dispose(curritem);
          curritem := nil;
        end;
  end;
  rootitem    := nil;
  curritem    := nil;
  curritemidx := -1;
  fcount      := 0;
end;
//-------------------------------------------------------------------
destructor TIfList.Destroy;
begin
  clear;
{$IFDEF CLASSLIST}
  inherited Destroy;
{$ENDIF}
end;
//-------------------------------------------------------------------
procedure TIfList.SetItem(Nr: LongInt; P: Pointer);
begin
  if not Assigned(RootItem) or (nr >= fCount) or (nr < 0) then
    Exit;
  if nr < curritemidx then
  begin
    CurrItem := RootItem;
    CurrItemIdx := 0;
  end;
  while curritemidx < nr do begin
    CurrItem := CurrItem^.Next;
    Inc(CurrItemIdx);
  end;
  CurrItem^.Ptr := p;
end;
//-------------------------------------------------------------------
function TifList.GetItem(Nr: LongInt): Pointer;
begin
  getitem := nil;
  if not Assigned(RootItem) or (nr >= fCount) or (nr < 0) then
    Exit;
  if nr < curritemidx then
  begin
    CurrItem := RootItem;
    CurrItemIdx := 0;
  end;
  while curritemidx < nr do begin
    CurrItem := CurrItem^.Next;
    Inc(CurrItemIdx);
  end;
  getitem := CurrItem^.Ptr;
end;
//-------------------------------------------------------------------
function TifList.Count: LongInt;
begin
 count := Fcount;
end;
//-------------------------------------------------------------------
function TIfStringList.Count: LongInt;
begin
  count := List.count;
end;
//-------------------------------------------------------------------
{$IFDEF SS}
type TsmallStr = record
    Size: Byte;
    Str: array[1..255] of Char;
  end;

function TifStringList.GetItem(Nr: LongInt): string;
var S: ^TSMALLSTR;
  tel: Byte;
begin
  getitem := '';
  s := List.GetItem(Nr);
  if s = nil then
    Exit;
  GetItem[0] := Chr(S^.Size);
  for tel := 1 to S^.Size do
    GetItem[tel] := S^.Str[tel];
end;

procedure TifStringList.SetItem(Nr: LongInt; const s: string);
var
  p: ^TSMALLSTR;
  tel: LongInt;
begin
  p := List.GetItem(Nr);
  if p = nil then
    Exit;
  FreeMem(p, p^.Size + 1);
  GetMem(p, Length(s) + 1);
  p^.Size := Length(s);
  for tel := 1 to p^.Size do
    p^.Str[tel] := s[tel];
  List.SetItem(Nr, p);
end;
//-------------------------------------------------------------------
procedure TifStringList.Add(const P: string);
var S: ^TSMALLSTR;
  tel: Byte;
begin
  GetMem(S, 1 + Length(P));
  s^.Size := Length(p);
  for tel := 1 to s^.Size do
    s^.Str[tel] := p[tel];
  List.Add(S);
end;
//-------------------------------------------------------------------
procedure TifStringList.Delete(NR: LongInt);
var P: ^TSMALLSTR;
begin
  p := list.getitem(nr);
  list.Remove(P);
  dispose(p);
end;

{$else
}
type pStr = ^string;

//-------------------------------------------------------------------
function TifStringList.GetItem(Nr: LongInt): string;
var
  S: PStr;
begin
  s := List.GetItem(Nr);
  if s = nil then
    Result := ''
  else

    Result := s^;
end;
//-------------------------------------------------------------------
procedure TifStringList.SetItem(Nr: LongInt; const s: string);
var
  p: PStr;
begin
  p := List.GetItem(Nr);
  if p = nil
    then
    Exit;
  p^ := s;
end;
//-------------------------------------------------------------------
procedure TifStringList.Add(const P: string);
var
  w: PStr;
begin
  new(w);
  w^ := p;
  List.Add(w);
end;
//-------------------------------------------------------------------
procedure TifStringList.Delete(NR: LongInt);
var
  W: PStr;
begin
  W := list.getitem(nr);
  if assigned(w) then
  begin
    dispose(w);
    list.Delete(Nr);
  end;
end;
{$ENDIF}

procedure TifStringList.Clear;
begin
  while List.Count > 0 do Delete(0);
end;

constructor TifStringList.Create;
begin
{$IFDEF CBUILDER}
  inherited Create;
  List := TList.Create;
{$ELSE}
  List.Create;
{$ENDIF}
end;

destructor TifStringList.Destroy;
var I: LongInt;
begin
  for I := 0 to list.count - 1 do
    Delete(I);
  List.Destroy;
{$IFDEF CLASSLIST}
  inherited Destroy;
{$ENDIF}
end;

//-------------------------------------------------------------------
function FastUpperCase(const s: string): string;
{Fast uppercase}
var
  i: Integer;
begin
  FastUpperCase := s;
  for i := 1 to Length(s) do
    if (s[i] >= #97) and (s[i] <= #122) then
      FastUpperCase[i] := char(byte(s[i]) and $DF);
end;
//-------------------------------------------------------------------
function FastLowerCase(const s: string): string;
{Fast lowercase}
var
  i: Integer;
begin
  FastLowerCase := s;
  for i := 1 to Length(s) do
    if (s[i] >= #65) and (s[i] <= #90) then
      FastLowerCase[i] := char(byte(s[i])+32);
end;
//-------------------------------------------------------------------

type
  TRTab = record
    name: string[20];
    c: TIfPasToken;
  end;

const
   KEYWORD_COUNT = 35;
const
  LookupTable: array[0..KEYWORD_COUNT] of TRTab = (
    (name: 'AND'; c: CSTII_and),
    (name: 'ARRAY'; c: CSTII_array),
    (name: 'BEGIN'; c: CSTII_begin),
    (name: 'CASE'; c: CSTII_case),
    (name: 'CONST'; c: CSTII_const),
    (name: 'DIV'; c: CSTII_div),
    (name: 'DO'; c: CSTII_do),
    (name: 'DOWNTO'; c: CSTII_downto),
    (name: 'ELSE'; c: CSTII_else),
    (name: 'END'; c: CSTII_end),
    (name: 'EXIT'; c : CSTII_exit),
    (name: 'FOR'; c: CSTII_for),
    (name: 'FUNCTION'; c: CSTII_function),
    (name: 'IF'; c: CSTII_if),
    (name: 'IN'; c: CSTII_in),
    (name: 'MOD'; c: CSTII_mod),
    (name: 'NOT'; c: CSTII_not),
    (name: 'OF'; c: CSTII_of),
    (name: 'OR'; c: CSTII_or),
    (name: 'PROCEDURE'; c: CSTII_procedure),
    (name: 'PROGRAM'; c: CSTII_program),
    (name: 'REPEAT'; c: CSTII_repeat),
    (name: 'SET'; c: CSTII_set),
    (name: 'SHL'; c: CSTII_shl),
    (name: 'SHR'; c: CSTII_shr),
    (name: 'STRING'; c: CSTII_string),
    (name: 'THEN'; c: CSTII_then),
    (name: 'TO'; c: CSTII_to),
    (name: 'TYPE'; c: CSTII_type),
    (name: 'UNTIL'; c: CSTII_until),
    (name: 'USES'; c: CSTII_uses),
    (name: 'VAR'; c: CSTII_var),
    (name: 'WHILE'; c: CSTII_while),
    (name: 'WITH'; c: CSTII_with),
    (name: 'XOR'; c: CSTII_xor),
    (name: 'BREAK'; c: CSTII_break));

//-------------------------------------------------------------------
function CheckReserved(data: PIfPascalParser): Boolean;
{Check if an identifier is a reserved word}
var
  s: string;
  i: Longint;
begin
  s := FastUppercase(GetToken(Data));
  for i := 0 to KEYWORD_COUNT do
     begin
        if s = LookupTable[i].Name then
        begin
           Data^.currTokenId := LookupTable[i].C;
           result := True;
           exit;
        end;
     end;
  result := False;
end;
//-------------------------------------------------------------------
procedure ParseToken(data: PIfPascalParser);
{Parse the token}
var
  ct, ci: LongInt;
  hs: Boolean;
begin

  ct := data^.CurrTokenPos;
  case data^.Text[ct] of
    #0:
      begin
        Data^.CurrTokenId := CSTI_EOF;
        Data^.CurrTokenLen := 0;
      end;
    'A'..'Z', 'a'..'z', '_':
      begin
        ci := ct + 1;
        while (data^.Text[ci] in ['_', '0'..'9', 'a'..'z', 'A'..'Z']) do begin
          Inc(ci);
        end;
        Data^.CurrTokenLen := ci - ct;
        if not CheckReserved(data) then
        begin
          Data^.CurrTokenId := CSTI_Identifier;
        end;
      end;
    '$':
      begin
        ci := ct + 1;

        while (data^.Text[ci] in ['0'..'9', 'a'..'f', 'A'..'F'])
        do Inc(ci);

        Data^.CurrTokenId  := CSTI_HexInt;
        Data^.CurrTokenLen := ci - ct;
      end;

    '0'..'9':
      begin
        hs := False;
        ci := ct;
        while (data^.Text[ci] in ['0'..'9']) do
        begin
          Inc(ci);
          if (data^.Text[ci] = '.') and (not hs) then
          begin
            hs := True;
            Inc(ci);
          end;
        end;

        if hs
        then Data^.CurrTokenId := CSTI_Real
        else Data^.CurrTokenId := CSTI_Integer;
        
        Data^.CurrTokenLen := ci - ct;
      end;
    #39:
      begin
        ci := ct + 1;
        while (data^.Text[ci] <> #0) and (data^.Text[ci] <> #13) and
          (data^.Text[ci] <> #10) and (data^.Text[ci] <> #39)
          do begin
          Inc(ci);
        end;
        if data^.Text[ci] = #39 then
          Data^.CurrTokenId := CSTI_String
        else

          Data^.CurrTokenId := CSTI_StringError;
        Data^.CurrTokenLen := ci - ct + 1;
      end;
    '#':
      begin
        ci := ct + 1;
        if data^.Text[ci] = '$' then
        begin
          while (data^.Text[ci] in ['A'..'Z', 'a'..'z', '0'..'9']) do begin
            Inc(ci);
          end;
          Data^.CurrTokenId := CSTI_Char;
          Data^.CurrTokenLen := ci - ct - 1;
        end else
        begin
          while (data^.Text[ci] in ['0'..'9']) do begin
            Inc(ci);
          end;
          if data^.Text[ci] in ['A'..'Z', 'a'..'z', '_'] then
            Data^.CurrTokenId := CSTI_CharError
          else

            Data^.CurrTokenId := CSTI_Char;
          Data^.CurrTokenLen := ci - ct;
        end;
      end;
    '=':
      begin
        Data^.CurrTokenId := CSTI_Equal;
        Data^.CurrTokenLen := 1;
      end;
    '>':
      begin
        if data^.Text[ct + 1] = '=' then
        begin
          Data^.CurrTokenid := CSTI_GreaterEqual;
          Data^.CurrTokenLen := 2;
        end else
        begin
          Data^.CurrTokenid := CSTI_Greater;
          Data^.CurrTokenLen := 1;
        end;
      end;
    '<':
      begin
        if data^.Text[ct + 1] = '=' then
        begin
          Data^.CurrTokenId := CSTI_LessEqual;
          Data^.CurrTokenLen := 2;
        end else
          if data^.Text[ct + 1] = '>' then
          begin
            Data^.CurrTokenId := CSTI_NotEqual;
            Data^.CurrTokenLen := 2;
          end else
          begin
            Data^.CurrTokenId := CSTI_Less;
            Data^.CurrTokenLen := 1;
          end;
      end;
    ')':
      begin
        Data^.CurrTokenId  := CSTI_CloseRound;
        Data^.CurrTokenLen := 1;
      end;
    '(':
      begin
        if Data^.Text[ct + 1] = '*' then
        begin
          ci := ct + 1;
          while (Data^.Text[ci] <> #0) do begin
            if (Data^.Text[ci] = '*') and (Data^.Text[ci + 1] = ')') then
              Break;
            Inc(ci);
          end;
          if (Data^.Text[ci] = #0) then
            Data^.CurrTokenId := CSTI_CommentEOFError
          else
          begin
            Data^.CurrTokenId := CSTI_Comment;
            Inc(ci, 2);
          end;
          Data^.CurrTokenLen := ci - ct;
        end
        else
        begin
          Data^.CurrTokenId := CSTI_OpenRound;
          Data^.CurrTokenLen := 1;
        end;
      end;
    '[':
      begin
        Data^.CurrTokenId  := CSTI_OpenBlock;
        Data^.CurrTokenLen := 1;
      end;
    ']':
      begin
        Data^.CurrTokenId  := CSTI_CloseBlock;
        Data^.CurrTokenLen := 1;
      end;
    ',':
      begin
        Data^.CurrTokenId  := CSTI_Comma;
        Data^.CurrTokenLen := 1;
      end;
    '.':
      begin
        Data^.CurrTokenId  := CSTI_Period;
        Data^.CurrTokenLen := 1;
      end;
    ';':
      begin
        Data^.CurrTokenId := CSTI_Semicolon;
        Data^.CurrTokenLen := 1;
      end;
    ':':
      begin
        if Data^.Text[ct + 1] = '=' then
        begin
          Data^.CurrTokenId  := CSTI_Assignment;
          Data^.CurrTokenLen := 2;
        end else
        begin
          Data^.CurrTokenId  := CSTI_Colon;
          Data^.CurrTokenLen := 1;
        end;
      end;
    '+':
      begin
        Data^.CurrTokenId  := CSTI_Plus;
        Data^.CurrTokenLen := 1;
      end;
    '-':
      begin
        Data^.CurrTokenId  := CSTI_Minus;
        Data^.CurrTokenLen := 1;
      end;
    '*':
      begin
        Data^.CurrTokenId  := CSTI_Multiply;
        Data^.CurrTokenLen := 1;
      end;
    '/':
      begin
        if data^.Text[ct + 1] = '/' then
        begin
          ci := ct + 1;
          while (Data^.Text[ci] <> #0) and (Data^.Text[ci] <> #13) and
            (Data^.Text[ci] <> #10) do begin
            Inc(ci);
          end;
          if (Data^.Text[ci] = #0) then
            Data^.CurrTokenId := CSTI_CommentEOFError
          else
          begin
            if Data^.Text[ci + 1] = #10 then
              Inc(ci) else

              if Data^.Text[ci + 1] = #13 then
                Inc(ci);
            Data^.CurrTokenId := CSTI_Comment;
          end;
          Data^.CurrTokenLen := ci - ct + 1;
        end else
        begin
          Data^.CurrTokenId := CSTI_Divide;
          Data^.CurrTokenLen := 1;
        end;
      end;
    #32, #9, #13, #10:
      begin
        ci := ct + 1;
        while (Data^.Text[ci] in [#32, #9, #13, #10]) do begin
          Inc(ci);
        end;
        Data^.CurrTokenId  := CSTI_Whitespace;
        Data^.CurrTokenLen := ci - ct;
      end;
    '{':
      begin
        ci := ct + 1;
        while (Data^.Text[ci] <> #0) and (Data^.Text[ci] <> '}') do begin
          Inc(ci);
        end;
        if (Data^.Text[ci] = #0) then
          Data^.CurrTokenId := CSTI_CommentEOFError
        else
          Data^.CurrTokenId := CSTI_Comment;
        Data^.CurrTokenLen := ci - ct + 1;
      end;
  else
    Data^.CurrTokenId := CSTI_SyntaxError;
    Data^.CurrTokenLen := 1;
  end;
end;
//-------------------------------------------------------------------
function GetToken(data: PIfPascalParser): string;
{Return the token}
var
  i: LongInt;
  len: LongInt;
  s: string;
begin
  len := data^.CurrTokenLen;
{$IFDEF SS}
  len  := len and $FF;
  s[0] := Chr(len);
{$ELSE}
  setlength(s, len);
{$ENDIF}
  for i := 1 to Len do begin
    s[i] := data^.Text[data^.CurrTokenPos + i - 1];
  end;
  GetToken := s;
end;
//-------------------------------------------------------------------
procedure SkipWhiteSpace(data: PIfPascalParser);
{Next no junk}
begin
  Data^.CurrTokenPos := Data^.CurrTokenPos + Data^.CurrTokenLen;
  ParseToken(data);
  while (Data^.CurrTokenId = CSTI_COMMENT) or (Data^.CurrTokenId =
    CSTI_WHITESPACE) do
  begin
    if (Data^.CurrTokenId = CSTI_EOF) or (Data^.CurrTokenId = CSTI_SyntaxError)
      or
      (Data^.CurrTokenId = CSTI_CommentEOFError) or (Data^.CurrTokenId =
      CSTI_CharError) or
      (Data^.CurrTokenId = CSTI_StringError)
      then
    begin
      Break;
    end;
    Data^.CurrTokenPos := Data^.CurrTokenPos + Data^.CurrTokenLen;
    ParseToken(data);
  end;
end;
//-------------------------------------------------------------------

end.

