{
IfPasScript
version: 2.29
Last stable version: 2.25

Parts ready:
 - Calculation
 - Standard types:  Byte, Shortint, Char, Word, SmallInt,
 - Standard types:  Cardinal, Longint, Integer, String, Real,
 - Standard types:  Double, Single, Extended, Comp, Boolean
 - Array(Only dynamic; 0 based; can be set using SetArrayLength(Array, Length))
 - Assignments (a:=b;)
 - If Then Else
 - For/To/Downto/Do
 - While/Begin/End
 - Case x Of/End
 - Repeat/Until
 - External Procedure/Function calls
 - Variable parameters for internal and extenal functions.
 - Internal Procedure/Functions
 - Internal Procedure calls from outside the script.
 - Standard library
 - Constants
 - Exit
 - Break
 - Uses (One uses can only be added once)
 - Documentation and examples

To do:
 - Records
 - Class support (internal)
 - Class support (external)
}
unit IfSPas;
{$I IfS_DEF.INC}
interface
uses
  IFS_VAR, IFS_UTL, math;
const
  Version = '2.29';

type
  TIfPasScript = class;
  TCs2PascalScript = TIfPasScript;
  TOnUses = function(ID: Pointer; Sender: TIfPasScript; Name: string): TCs2Error;
  TOnRunLine = function(ID: Pointer; Sender: TIfPasScript; Position: Longint): TCs2Error;
  TResourceFree = procedure(ID: Pointer; Data: Pointer);

  TIfPasScript = class
  private
    FUses: TifStringList;
    FUnits: TifList;
    InternalProcedures: PProcedureManager;
    FAllocatedResources: TifList;

    Text: PChar;
    MainOffset: LongInt;
    FId: Pointer;
    Parser: PIfPascalParser;
    FErrorPos: LongInt;
    FErrorCode: TCs2Error;
    FOnUses: TOnUses;
    FOnRunLine: TOnRunLine;
    Procedures: PProcedureManager;
    function CalcSub(Vars: PVariableManager; var W: PIfVariant): Boolean;
    function IdentifierExists(SubVars: PVariableManager; const S: string): Boolean;
    function ProcessVars(Vars: PVariableManager): Boolean;
    function ProcessConsts(Vars: PVariableManager): Boolean;
    procedure RunError(C: TCs2Error);
    function ReadType: PTypeRec;
    function RunBegin(Vars: PVariableManager; Skip: Boolean): Boolean;
    function Calc(Vars: PVariableManager; res: PIfVariant;
      StopOn: TIfPasToken; OnlyConst: Boolean): Boolean;
    function DoProc(Vars: PVariableManager; Internal: Boolean): PIfVariant;
  public
    Variables: PVariableManager;
    Types: PTypeManager;
    function AddFunction(Proc: Pointer; Decl: String): Boolean;
    property GetText: Pchar read Text;
    property OnRunLine: TOnRunLine read FOnRunLine write FOnRunLine;
    property OnUses: TOnUses read FOnUses write FOnUses;
    property ErrorCode: TCs2Error read FErrorCode;
    property ErrorPos: LongInt read FErrorPos;

    procedure RunScript;
    function RunScriptProc(const Name: string; Parameters: PVariableManager): PIfVariant;

    procedure AddResource(FreeProc: TResourceFree; Data: Pointer);
    procedure RemoveResource(Data: Pointer);
    function IsValidResource(FreeProc: TResourceFree; Data: Pointer): Boolean;

    procedure SetText(p: Pchar);
    constructor Create(ID: Pointer);
    destructor Destroy; override;
  end;

procedure RegisterStdLib(P: TIfPasScript);
{Register all standard functions:}
{
Install:
  Function StrGet(S : String; I : Integer) : Char;
  Function StrSet(c : Char; I : Integer; var s : String) : Char;
  Function Ord(C : Char) : Byte;
  Function Chr(B : Byte) : Char;
  Function StrToInt(s : string;def : Longint) : Longint;
  Function IntToStr(i : Longint) : String;
  Function Uppercase(s : string) : string;
  Function Lowercase(s : string) : string;
  Function Trim(s : string) : string;
  Function Copy(S : String; Indx, Count : Integer) : String;
  Procedure Delete(var S : String; Indx, Count : Integer);
  Function Pos(SubStr, S : String) : Integer;
  Procedure Insert(Source : String; var Dest : String; Indx : Integer);
  Function Length(s : String) : Longint;
  Function Sin(e : Extended) : Extended;
  Function Cos(e : Extended) : Extended;
  Function Sqrt(e : Extended) : Extended;
  Function Round(e : Extended) : Longint;
  Function Trunc(e : Extended) : Longint;
  Function Int(e : Extended) : Longint;
  Function Pi : Extended;
  Function Abs(e : Extended) : Extended;
  Function Sqrt(e : Extended) : Extended;
  Function FloatToStr(e : Extended) : String;
  Function Padl(s : string;I : longInt) : string;
  Function Padr(s : string;I : longInt) : string;
  Function Padz(s : string;I : longInt) : string;
  Function Replicate(c : char;I : longInt) : string;
  Function StringOfChar(c : char;I : longInt) : string;
  procedure SetArrayLength(var u: array; Length: Longint);
  function GetArrayLengtH(var u: array): Longint;
}
function CreateReal(FTypes: PTypeManager; const e: Extended): PIfVariant;
function CreateString(FTypes: PTypeManager; const s: string): PIfVariant;
function CreateInteger(FTypes: PTypeManager; i: LongInt): PIfVariant;
function CreateBool(FTypes: PTypeManager; b: Boolean): PIfVariant;

implementation

function CreateReal(FTypes: PTypeManager; const e: Extended): PIfVariant;
var
  p: PIfVariant;
begin
  p := CreateCajVariant(TM_Add(FTypes, '', CSV_Extended, nil));
  p^.Cv_Extended := e;
  CreateReal := p;
end;

function CreateString(FTypes: PTypeManager; const s: string): PIfVariant;
var
  p: PIfVariant;
begin
  p := CreateCajVariant(TM_Add(FTypes, '', CSV_String, nil));
  p^.Cv_Str := s;
  CreateString := p;
end;

function CreateInteger(FTypes: PTypeManager; i: LongInt): PIfVariant;
var
  p: PIfVariant;
begin
  p := CreateCajVariant(TM_Add(FTypes, '', CSV_SInt32, nil));
  p^.Cv_sInt32 := i;
  CreateInteger := p;
end;

function CreateBool(FTypes: PTypeManager; b: Boolean): PIfVariant;
var
  p: PIfVariant;
begin
  p := CreateCajVariant(TM_Add(FTypes, '', CSV_Bool, nil));
  p^.Cv_Bool := b;
  Createbool := p;
end;



type
  PResourceData = ^TResourceData;
  TResourceData = record
    Data: Pointer;
    FreeProc: TResourceFree;
  end;
  PUnit = ^TUnit;
  TUnit = record
    Name: string;
    DataLen: Longint;
    Data: Pointer;
    Procs: PProcedureManager;
  end;

procedure AddStandard(p: PTypeManager);
begin
  TM_Add(p, 'BYTE', CSV_UByte, nil);
  TM_Add(p, 'SHORTINT', CSV_SByte, nil);
  TM_Add(p, 'CHAR', CSV_Char, nil);
  TM_Add(p, 'WORD', CSV_UInt16, nil);
  TM_Add(p, 'SMALLINT', CSV_SInt16, nil);
  TM_Add(p, 'CARDINAL', CSV_UInt32, nil);
  TM_Add(p, 'LONGINT', CSV_SInt32, nil);
  TM_Add(p, 'INTEGER', CSV_SInt32, nil);
  TM_Add(p, 'STRING', CSV_String, nil);
  TM_Add(p, 'REAL', CSV_Real, nil);
  TM_Add(p, 'SINGLE', CSV_Single, nil);
  TM_Add(p, 'DOUBLE', CSV_Double, nil);
  TM_Add(p, 'EXTENDED', CSV_Extended, nil);
  TM_Add(p, 'COMP', CSV_Comp, nil);
  TM_Add(p, 'BOOLEAN', CSV_Bool, nil);
  TM_Add(p, 'ARRAY', CSV_Array, nil); // only for internal use! (SetArrayLength; GetArrayLength)
end;

//-------------------------------------------------------------------

function IntToStr(I: LongInt): string;
var
  s: string;
begin
  Str(i, s);
  IntToStr := s;
end;
//-------------------------------------------------------------------

function FloatToStr(E: Extended): string;
var
  s: string;
begin
  Str(e: 12: 2, s);
  result := s;
end;
//-------------------------------------------------------------------

function Padl(s: string; i: longInt): string;
var
  x: integer;
begin
  result := s;
  x := length(result);
  if x < i then
    result := StringOfChar(' ', i - x) + result;
end;
//-------------------------------------------------------------------

function Padz(s: string; i: longInt): string;
var
  x: integer;
begin
  result := s;
  x := length(result);
  if x < i then
    result := StringOfChar('0', i - x) + result;
end;
//-------------------------------------------------------------------

function Padr(s: string; i: longInt): string;
begin
  result := s + StringOfChar(' ', i);
  result := copy(result, 1, i);
end;
//-------------------------------------------------------------------

function FastTrim(s: string): string;
var
  x, y: integer;
begin
  result := '';
  y := length(s);
  for x := 1 to y do
    if s[x] <> #32
      then result := result + s[x];
end;
//-------------------------------------------------------------------

function StrToInt(const S: string): LongInt;
var
  e: Integer;
  Res: LongInt;
begin
  Val(S, Res, e);
  if e <> 0 then
    StrToInt := -1
  else
    StrToInt := Res;
end;
//-------------------------------------------------------------------

function StrToIntDef(const S: string; Def: LongInt): LongInt;
var
  e: Integer;
  Res: LongInt;
begin
  Val(S, Res, e);
  if e <> 0 then
    StrToIntDef := Def
  else
    StrToIntDef := Res;
end;
//-------------------------------------------------------------------

function StrToReal(const S: string): Extended;
var
  e: Integer;
  Res: Extended;
begin
  Val(S, Res, e);
  if e <> 0 then
    StrToReal := -1
  else
    StrToReal := Res;
end;
//-------------------------------------------------------------------

function IntProcDefParam(S: string; I: Integer): Integer;
{
Parse the incode-script procedure definition from a string.
When I=0 this function will return the result type.
When I=-1 this function will return the number of parameters.
When I=1 this function will return the first parameter type.
When I=2 this function will return the second parameter type.
etc.
}
var
  Res: Integer;
begin
  if I = 0 then
  {Return result-type} IntProcDefParam := StrToInt(Fw(s)) else
    if I = -1 then
    {Return param count} begin
      res := 0;
      Delete(S, 1, Length(Fw(s))); {result}
      Rs(S);
      Delete(S, 1, Length(Fw(s))); {name}
      Rs(S);
      while Length(s) > 0 do begin
        Inc(Res);
        Delete(S, 1, Length(Fw(s))); {Delete parameter name}
        Rs(S);
        Delete(S, 1, Length(Fw(s))); {Delete parameter type}
        Rs(S);
      end; {while}
      IntProcDefParam := Res;
    end {else if} else begin
      res := 0;
      if I < 1 then
      begin IntProcDefParam := -1; Exit; end;
      Delete(S, 1, Length(Fw(s))); {result}
      Rs(S);
      Delete(S, 1, Length(Fw(s))); {name}
      Rs(S);
      while Length(s) > 0 do begin
        Inc(Res);
        Delete(S, 1, Length(Fw(s))); {delete parameter name}
        Rs(S);
        if Res = I then
        begin IntProcDefParam := StrToInt(Fw(s)); Exit; end;
        Delete(S, 1, Length(Fw(s))); {delete type}
        Rs(S);
      end; {while}
      IntProcDefParam := 0;
    end {Else Else if}
end; {IntProcDefParam}
//-------------------------------------------------------------------

function IntProcDefName(S: string; I: Integer): string;
{
Parse the incode-script procedure definition from a string.
i=0 will return the procedure name
I=1 will return the first one
}
var
  Res: Integer;
begin
  res := 0;
  if i = 0 then
  begin
    Delete(S, 1, Length(Fw(s))); {result}
    Rs(S);
    IntProcDefName := fw(s);
    Exit;
  end;
  if I < 1 then
  begin IntProcDefName := ''; Exit; end;
  Delete(S, 1, Length(Fw(s))); {result}
  Rs(S);
  Delete(S, 1, Length(Fw(s))); {name}
  Rs(S);
  while Length(s) > 0 do begin
    Inc(Res);
    if Res = I then
    begin IntProcDefName := Fw(s); Exit; end;
    Delete(S, 1, Length(Fw(s))); {delete parameter name}
    Rs(S);
    Delete(S, 1, Length(Fw(s))); {delete type}
    Rs(S);
  end; {while}
  IntProcDefName := '';
end; {IntProcDefParam}
//-------------------------------------------------------------------

function TIFPasScript.ReadType: PTypeRec;
var
  Ex: Pointer;
begin
  if Parser^.CurrTokenId = CSTII_Array then
  begin
    SkipWhiteSpace(Parser);
    if Parser^.CurrTokenId <> CSTII_Of then
    begin
      RunError(EOfExpected);
      ReadType := nil;
      Exit;
    end;
    SkipWhiteSpace(Parser);
    Ex := ReadType;
    if ex <> nil then
      ReadType := TM_Add(Types, '', CSV_Array, Ex)
    else
      ReadType := nil;
  end else begin
    ReadType := TM_Get(Types, FastUpperCase(GetToken(Parser)));
  end;
end;

//-------------------------------------------------------------------

function TIfPasScript.CalcSub(Vars: PVariableManager; var W: PIfVariant): Boolean;
{Calculate array [integer]}
var
  r: PIfVariant;
  op: Longint;
  InOpenBlock: Boolean;
begin
  CalcSub := False;
  op := Parser.CurrTokenPos;
  SkipWhiteSpace(Parser);
  InOpenBlock := False;
  while True do begin
    if w.VType.atypeid <> CSV_Array then Break;
    if (Parser^.CurrTokenId <> CSTI_OpenBlock) and not InOpenBlock then Break;
    SkipWhiteSpace(Parser);
    r := CreateCajVariant(TM_Add(Types, '', CSV_SInt32, nil));
    if not Calc(Vars, r, CSTI_CloseBlock, False) then
    begin
      DestroyCajVariant(r);
      Exit;
    end; {if}
    if Parser^.CurrTokenID = CSTI_Comma then
    begin
      InOpenBlock := True;
    end else
    if Parser^.CurrTokenID <> CSTI_CloseBlock then
    begin
      RunError(ECloseBlockExpected);
      DestroyCajVariant(r);
      Exit;
    end else InOpenBlock := False;
    w := w^.CV_ArrItems.GetItem(r^.CV_SInt32); 
    DestroyCajVariant(r);
    if not Assigned(w) then
    begin
      RunError(EOutOfRange);
      Exit;
    end; {if}
    op := Parser.CurrTokenPos;
    if not InOpenBlock then
      SkipWhiteSpace(Parser);
  end;
  Parser.CurrTokenPos := Op;
  ParseToken(Parser);
  if InOpenBlock then
  begin
    RunError(ECloseBlockExpected);
  end;
  CalcSub := True;
end; {CalcArrayInt}
//-------------------------------------------------------------------

function TIfPasScript.IdentifierExists(SubVars: PVariableManager; const S: string): Boolean;
{ Check if an identifier exists }
  function UsesExists(s: string): Boolean;
  var
    i: Integer;
  begin
    UsesExists := False;
    for i := 0 to FUses.Count - 1 do
      if FUses.GetItem(i) = s then
      begin
        UsesExists := True;
        Break;
      end;
  end; { UsesExists }

begin
  IdentifierExists := False;
  if UsesExists(FastUppercase(s)) then
    IdentifierExists := True
  else if PM_Find(Procedures, FastUppercase(s)) <> -1 then
    IdentifierExists := True
  else if PM_Find(InternalProcedures, FastUppercase(s)) <> -1 then
    IdentifierExists := True
  else if VM_Find(Variables, FastUppercase(s)) <> -1 then
    IdentifierExists := True
  else if TM_Get(Types, FastUppercase(s)) <> nil then
    IdentifierExists := True
  else if Assigned(SubVars) and (VM_Find(subVars, FastUppercase(s)) <> -1) then
  begin
    IdentifierExists := True
  end;
end; {IdentifierExists}

//-------------------------------------------------------------------

procedure TIfPasScript.SetText(p: PChar);
{ Assign a text to the script engine, this also checks for uses and variables. }
var
  HaveHadProgram,
    HaveHadUses: Boolean;

  function ProcessUses: Boolean;
  {Process Uses block}
  var
    i: Integer;
  begin
    ProcessUses := False;
    while Parser^.CurrTokenId <> CSTI_EOF do begin
      if Parser^.CurrTokenId <> CSTI_Identifier then
      begin
        RunError(EIdentifierExpected);
        Exit;
      end; {If}
      if IdentifierExists(nil, GetToken(Parser)) then
      begin
        RunError(EDuplicateIdentifier);
        Exit;
      end; {If}
      FUses.Add(FastUpperCase(GetToken(Parser)));
      if Assigned(OnUses) then
      begin
        i := OnUses(fId, Self, GetToken(Parser));
        if I <> ENoError then
        begin
          RunError(i);
          Exit;
        end; {If}
      end {If}
      else
      begin
        RunError(EUnknownIdentifier);
        Exit;
      end; {Else if}
      SkipWhiteSpace(Parser);
      if (Parser^.CurrTokenId = CSTI_SemiColon) then
      begin
        SkipWhiteSpace(Parser);
        Break;
      end {if}
      else if (Parser^.CurrTokenId <> CSTI_Comma) then
      begin
        RunError(EDuplicateIdentifier);
        Exit;
      end else {Else if}
      begin
        SkipWhiteSpace(Parser);
      end;
    end;
    if Parser^.CurrTokenId = CSTI_EOF then
    begin
      RunError(EUnexpectedEndOfFile);
    end {If}
    else
    begin
      ProcessUses := True;
    end; {Else If}
  end; {ProcessUses}

  function DoFuncHeader: Boolean;
  var
    FuncParam: string;
    FuncName: string;
    CurrVar: string;
    CurrType: Pointer;
    FuncRes: Pointer;
    function Duplic(S: string): Boolean;
    var
      s2, s3: string;
      i: Integer;
    begin
      if s = FuncName then
      begin
        Duplic := True;
        Exit;
      end; {if}
      if (funcRes <> nil) and (s = 'RESULT') then
      begin
        duplic := True;
        Exit;
      end;
      s2 := CurrVar;
      while Pos('|', s2) > 0 do begin
        if Pos('!', s2) = 1 then
          Delete(s2, 1, 1);
        if Copy(s2, 1, Pos('|', s2) - 1) = s then
        begin
          Duplic := True;
          Exit;
        end; {if}
        Delete(s2, 1, Pos('|', s2));
      end; {while}
      s2 := '0 ' + FuncParam;
      for i := 1 to IntProcDefParam(s2, -1) do begin
        s3 := IntProcDefName(s2, 0);
        if Pos('!', s2) = 1 then
          Delete(s2, 1, 1);
        if s3 = s then
        begin
          Duplic := True;
          Exit;
        end; {if}
      end; {for}
      Duplic := False;
    end; {duplic}
  begin
    DoFuncHeader := False;
    if Parser^.CurrTokenId = CSTII_Procedure then
      FuncRes := nil
    else
      FuncRes := pointer(1);
    SkipWhiteSpace(Parser);
    if Parser^.CurrTokenId <> CSTI_Identifier then
    begin
      RunError(EIdentifierExpected);
      Exit;
    end; {if}
    if IdentifierExists(nil, GetToken(Parser)) then
    begin
      RunError(EDuplicateIdentifier);
      Exit;
    end; {if}
    FuncName := FastUppercase(GetToken(Parser));
    FuncParam := FuncName;
    CurrVar := '';
    SkipWhiteSpace(Parser);
    if parser^.CurrTokenId = CSTI_OpenRound then
    begin
      SkipWhiteSpace(Parser);
      if Parser^.CurrTokenID = CSTI_CloseRound then
      begin
        SkipWhiteSpace(Parser);
      end else
      begin
        while True do begin
          if Parser^.CurrTokenId = CSTII_Var then
          begin
            CurrVar := '!';
            SkipWhiteSpace(Parser);
          end; {if}
          while True do begin
            if Parser^.CurrTokenId <> CSTI_Identifier then
            begin
              RunError(EIdentifierExpected);
              Exit;
            end; {if}
            if IdentifierExists(nil, GetToken(Parser)) or Duplic(GetToken(Parser))
              then
            begin
              RunError(EDuplicateIdentifier);
              Exit;
            end; {if}
            CurrVar := CurrVar + fastuppercase(GetToken(Parser)) + '|';
            SkipWhiteSpace(parser);
            if Parser^.CurrTokenId = CSTI_Colon then
              Break;
            if Parser^.CurrTokenId <> CSTI_Comma then
            begin
              RunError(ECommaExpected);
              Exit;
            end; {if}
            SkipWhiteSpace(Parser);
          end; {while}
          SkipWhiteSpace(Parser);
          CurrType := TM_Get(Types, FastUppercase(GetToken(Parser)));
          if CurrType = nil then
          begin
            RunError(EUnknownIdentifier);
            Exit;
          end; {if}
          if Pos('!', CurrVar) = 1 then
          begin
            Delete(currVar, 1, 1);
            while Pos('|', CurrVar) > 0 do begin
              FuncParam := FuncParam + ' !' + Copy(CurrVar, 1, Pos('|', CurrVar) -
                1) + ' ' + IntToStr(Longint(CurrType));
              Delete(CurrVar, 1, Pos('|', CurrVar));
            end; {while}
          end else
          begin
            while Pos('|', CurrVar) > 0 do begin
              FuncParam := FuncParam + ' ' + Copy(CurrVar, 1, Pos('|', CurrVar) -
                1) + ' ' + IntToStr(Longint(CurrType));
              Delete(CurrVar, 1, Pos('|', CurrVar));
            end; {while}
          end; {if}
          SkipWhiteSpace(Parser);
          if Parser^.CurrTokenId = CSTI_CloseRound then
          begin
            SkipWhiteSpace(Parser);
            Break;
          end; {if}
          if Parser^.CurrTokenId <> CSTI_SemiColon then
          begin
            RunError(ESemiColonExpected);
            Exit;
          end; {if}
          SkipWhiteSpace(Parser);
        end; {while}
      end; {else if}
    end; {if}
    if FuncRes <> nil then
    begin
      if Parser^.CurrTokenId <> CSTI_Colon then
      begin
        RunError(EColonExpected);
        Exit;
      end;
      SkipWhiteSpace(Parser);
      FuncRes := TM_GET(Types, FastUppercase(GetToken(Parser)));
      if FuncRes = nil then
      begin
        RunError(EUnknownIdentifier);
        Exit;
      end;
      SkipWhiteSpace(parser);
    end;
    FuncParam := InttoStr(Longint(FuncRes)) + ' ' + FuncParam;
    if Parser^.CurrTokenId <> CSTI_Semicolon then
    begin
      RunError(ESemiColonExpected);
      Exit;
    end;
    SkipWhiteSpace(Parser);
    PM_Add(InternalProcedures, FuncParam, Pointer(Parser^.CurrTokenPos));
    DoFuncHeader := True;
    if Parser^.CurrTokenId = CSTII_Var then
    begin
      while (Parser^.CurrTokenID <> CSTII_Begin) and (Parser^.CurrTokenID <>
        CSTI_EOF) do
        SkipWhiteSpace(Parser);
    end;
    RunBegin(nil, True);
    if Parser^.CurrTokenId <> CSTI_Semicolon then
    begin
      RunError(ESemiColonExpected);
      Exit;
    end;
    SkipWhiteSpace(Parser);
  end; {DoFuncHeader}

  function ProcessTypes: Boolean;
  var
    Name: String;
    p: PTypeRec;
  begin
    ProcessTypes := False;
    SkipWhiteSpace(Parser);
    repeat
      if Parser^.CurrTokenId <> CSTI_Identifier then
      begin
        RunError(EIdentifierExpected);
        Exit;
      end; {if}
      Name := FastUpperCase(GetToken(Parser));
      if IdentifierExists(nil, Name) then
      begin
        RunError(EDuplicateIdentifier);
        Exit;
      end; {if}
      SkipWhiteSpace(Parser);
      if Parser^.CurrTokenId <> CSTI_Equal then
      begin
        RunError(EIsExpected);
        Exit;
      end;
      SkipWhiteSpace(Parser);
      p := ReadType;
      if p = nil then
      begin
        RunError(EUnknownIdentifier);
        exit;
      end;
      SkipWhiteSpace(Parser);
      if Parser^.CurrTokenID <> CSTI_SemiColon then
      begin
        RunError(ESemicolonExpected);
        exit;
      end;
      SkipWhiteSpace(Parser);
      TM_Add(Types, Name, p^.atypeid, p^.ext);
    until Parser^.CurrTokenId <> CSTI_Identifier;
    ProcessTypes := True;
  end; {ProcessTypes}

begin
  FUses.Clear;
  VM_Clear(Variables);
  TM_Destroy(Types); Types := TM_Create;
  Vm_Add(Variables, CreateBool(Types, True), 'TRUE')^.Flags := 1;
  Vm_Add(Variables, CreateBool(Types, False), 'FALSE')^.Flags := 1;
  PM_Clear(Procedures);
  PM_Clear(InternalProcedures);
  AddStandard(Types);
  FUses.Add('SYSTEM');
  if Assigned(OnUses) then
    OnUses(fId, Self, 'SYSTEM');

  RunError(ENoError);
  MainOffset := -1;
  Text := p;
  if Text = nil then
  begin
    Exit;
  end; {If}
  Parser^.Text := Text;
  Parser^.CurrTokenPos := 0;
  HaveHadProgram := False;
  HaveHadUses := False;
  ParseToken(Parser);
  if (Parser^.CurrTokenId = CSTI_WhiteSpace) or (Parser^.CurrTokenId =
    CSTI_Comment) then
    SkipWhiteSpace(Parser);
  while Parser^.CurrTokenId <> CSTI_EOF do begin
    case Parser^.CurrTokenId of
      CSTI_CommentEOFError,
        CSTI_CharError,
        CSTI_SyntaxError: begin RunError(ESyntaxError); Exit; end;
      CSTI_StringError: begin RunError(EStringError); Exit; end;
    end;
    if (Parser^.CurrTokenId = CSTII_Program) and (HaveHadProgram = False) and
      (HaveHadUses = False) then
    begin
      SkipWhiteSpace(Parser);
      if Parser^.CurrTokenId <> CSTI_Identifier then
      begin
        RunError(EIdentifierExpected);
        Exit;
      end; {if}
      SkipWhiteSpace(Parser);
      if Parser^.CurrTokenId <> CSTI_Semicolon then
      begin
        RunError(ESemicolonExpected);
        Exit;
      end; {if}
      SkipWhiteSpace(Parser);
      HaveHadProgram := True;
    end {if}
    else if (Parser^.CurrTokenId = CSTII_Uses) and (HaveHadUses = False) then
    begin
      SkipWhiteSpace(Parser);
      if not ProcessUses then
        Exit;
      HaveHadUses := True;
    end {else if}
    else if (Parser^.CurrTokenId = CSTII_Type) then
    begin
      if not ProcessTypes then
        Exit;
    end
    else if (Parser^.CurrTokenId = CSTII_Var) then
    begin
      if not ProcessVars(Variables) then
        Exit;
    end {Else if}
    else if (Parser^.CurrTokenId = CSTII_Const) then
    begin
      if not ProcessConsts(Variables) then
        Exit;
    end
    else if (Parser^.CurrTokenId = CSTII_Procedure) or
      (Parser^.CurrTokenId = CSTII_Function)
      then
    begin
      if not DoFuncHeader then
        Exit;
    end {else if}
    else if (Parser^.CurrTokenId = CSTII_Begin) then
    begin
      MainOffset := Parser^.CurrTokenPos;
      Exit;
    end {Else if}
    else if (Parser^.CurrTokenId = CSTI_EOF) then
    begin
      RunError(EUnexpectedEndOfFile);
    end {Else if}
    else
    begin
      RunError(EBeginExpected);
      Exit;
    end; {Else If}
  end; {While}
end; {SetText}

//-------------------------------------------------------------------

function TIfPasScript.ProcessConsts(Vars: PVariableManager): Boolean;
        { Process constants block (const s = '') }
var
  Name: string;
  value: PIfVariant;
begin
  ProcessConsts := False;
  SkipWhiteSpace(Parser);
  repeat
    if parser^.CurrTokenId <> CSTI_Identifier then begin
      RunError(EIdentifierExpected);
      Exit;
    end;
    Name := FastUpperCase(GetToken(Parser));
    if IdentifierExists(Vars, Name) then begin
      RunError(EDuplicateIdentifier);
      exit;
    end;
    SkipWhiteSpace(Parser);
    if Parser^.CurrTokenID <> CSTI_Equal then begin
      RunError(EIsExpected);
      exit;
    end;
    SkipWhiteSpace(Parser);
    Value := CreateCajVariant(TM_Add(Types, '', CSV_Var, nil));
    Value^.CV_Var := nil;
    if not Calc(Vars, Value, CSTI_SemiColon, True) then begin
      DestroyCajVariant(value);
      exit;
    end;
    value^.Flags := $1;
    VM_Add(Vars, Value, name);
    SkipWhiteSpace(Parser);
  until Parser^.CurrTokenID <> CSTI_Identifier;
  ProcessConsts := True;
end;
//-------------------------------------------------------------------

function TIfPasScript.ProcessVars(Vars: PVariableManager): Boolean;
        { Process Vars block }
var
  Names: string;
  n: PTypeRec;
begin
  SkipWhiteSpace(Parser);
  Names := '';
  ProcessVars := False;
  while True do begin
    case Parser^.CurrTokenId of
      CSTI_CommentEOFError,
        CSTI_CharError,
        CSTI_SyntaxError: begin RunError(ESyntaxError); Exit; end;
      CSTI_StringError: begin RunError(EStringError); Exit; end;
      CSTI_EOF: begin RunError(EUnexpectedEndOfFile); Exit; end;
    end;
    if Parser^.CurrTokenId <> CSTI_Identifier then
    begin
      RunError(EIdentifierExpected);
      Exit;
    end;
    if IdentifierExists(Vars, GetToken(Parser)) then
    begin
      RunError(EDuplicateIdentifier);
      Exit;
    end; {if}

    Names := Names + FastUpperCase(GetToken(Parser)) + '|';
    SkipWhiteSpace(Parser);
    while Parser^.CurrTokenId = CSTI_Comma do begin
      SkipWhiteSpace(Parser);
      if Parser^.CurrTokenId <> CSTI_Identifier then
      begin
        RunError(EIdentifierExpected);
        Exit;
      end; {if}
      if IdentifierExists(nil, GetToken(Parser)) then
      begin
        RunError(EDuplicateIdentifier);
        Exit;
      end; {if}
      Names := Names + FastUpperCase(GetToken(Parser)) + '|';
      SkipWhiteSpace(Parser);
    end; {while}
    if Parser^.CurrTokenId <> CSTI_Colon then
    begin
      RunError(EColonExpected);
      Exit;
    end; {if}
    SkipWhiteSpace(Parser);
    n := ReadType;
    if N = nil then
    begin
      Exit;
    end; {if}
    while Pos('|', names) > 0 do begin
      VM_Add(Vars, CreateCajVariant(n), Copy(names, 1, Pos('|', names) - 1));
      Delete(Names, 1, Pos('|', Names));
    end; {if}
    SkipWhiteSpace(Parser);
    if Parser^.CurrTokenId <> CSTI_Semicolon then
    begin
      RunError(ESemicolonExpected);
      Exit;
    end; {if}
    SkipWhiteSpace(Parser);
    if Parser^.CurrTokenId <> CSTI_Identifier then
      Break;
  end; {while}
  ProcessVars := True;
end; {ProcessVars}

//-------------------------------------------------------------------

constructor TIfPasScript.Create(ID: Pointer);
begin
  inherited Create;
  fid := Id;
{$IFDEF CLASSLIST}
  FUses := TIfStringList.Create;
  FAllocatedResources := TIfList.Create;
  FUnits := TifList.Create;
{$ELSE}
  FUses.Create;
  FAllocatedResources.Create;
  FUnits.Create;
{$ENDIF}
  New(Parser);
  RunError(ENoError);
  Text := nil;
  MainOffset := -1;
  Procedures := PM_Create;
  InternalProcedures := PM_Create;
  Variables := VM_Create(nil);
  Types := TM_Create;
  AddStandard(Types);
  OnUses := nil;
  OnRunLine := nil;
end; {Create}
//-------------------------------------------------------------------

destructor TIfPasScript.Destroy;
var
  i: Longint;
  c: PUnit;
begin
  Dispose(Parser);
  VM_Destroy(Variables);
  PM_Destroy(InternalProcedures);
  PM_Destroy(Procedures);
  TM_Destroy(Types);
  FAllocatedResources.Destroy;
  FUses.Destroy;
  FUnits.Destroy;
  for i := 0 to FUnits.Count - 1 do begin
    c := FUnits{$IFDEF CLASSLIST} [i]{$ELSE}.GetItem(I){$ENDIF};
    PM_Destroy(C^.Procs);
    FreeMem(C^.Data, c^.DataLen);
    Dispose(C);
  end;
  inherited Destroy;
end; {Create}
//-------------------------------------------------------------------

procedure TIfPasScript.RunError(C: TCs2Error);
{ Run an error }
begin
  if c = ENoError then
  begin
    FErrorCode := C;
    FErrorPos := -1;
  end {if}
  else
  begin
    FErrorCode := C;
    FErrorPos := Parser^.CurrTokenPos;
  end; {else if}
end; {RunError}

//-------------------------------------------------------------------
// Procedure: RunScript
//   Purpose: Process the script commands
//-------------------------------------------------------------------

procedure TIfPasScript.RunScript;
var
  i: Longint;
  p: PResourceData;
begin
  if MainOffset = -1 then Exit;

  RunError(ENoError); // Reset the error code and position
  Parser^.CurrTokenPos := MainOffset; // Position for the next token
  ParseToken(Parser); // Get the token
  //
  // RunBegin actually parses the script and expects a final period
  //-----------------
  if RunBegin(nil, False) then
  begin
    if Parser^.CurrTokenId <> CSTI_Period
      then RunError(EPeriodExpected);
  end;
  if FErrorCode = EExitCommand then
    FErrorCode := 0;

  for i := 0 to FAllocatedResources.Count - 1 do
  begin
    p := FAllocatedResources.GetItem(i);
    p^.FreeProc(fId, P^.Data);
  end;
  FAllocatedResources.Clear;
end; {RunScript}
//-------------------------------------------------------------------
type
  PSmallCalculation = ^TSmallCalculation;
  TSmallCalculation = packed record
    TType: Byte;
                                  {
                                  0 = Variant

                                  2 = *
                                  3 = /
                                  4 = DIV
                                  5 = MOD
                                  6 = AND
                                  7 = SHR
                                  8 = SHL

                                  9 = +
                                  10 = -
                                  11 = OR
                                  12 = XOR

                                  13 = =
                                  14 = >
                                  15 = <
                                  16 = <>
                                  17 = <=
                                  18 = >=
                                  }
    CajVariant: PIfVariant;
  end;

function TIfPasScript.Calc(Vars: PVariableManager; res: PIfVariant; StopOn: 
  TIfPasToken; OnlyConst: Boolean): Boolean;
{ Calculate an expression }
var
  Items: TifList;
  PreCalc: string;
  temp4: PIfVariant;
  Work: PSmallCalculation;
  function ChrToStr(s: string): Char;
    {Turn a char intto a string}
  begin
    Delete(s, 1, 1); {First char : #}
    ChrToStr := Chr(StrToInt(s));
  end;
  function PString(s: string): string;
    { remove the ' from the strings}
  begin
    s := Copy(s, 2, Length(s) - 2);
    PString := s;
  end;
  function DoPrecalc: Boolean;
    {Pre calculate (- not +)}
  begin
    DoPrecalc := True;
    while Length(Precalc) > 0 do begin
      if precalc[1] = '-' then
      begin
        if not DoMinus(Work^.CajVariant) then
        begin
          RunError(ETypeMismatch);
          Exit;
        end;
      end else if precalc[1] = '|' then
      begin
        if not DoNot(Work^.CajVariant) then
        begin
          RunError(ETypeMismatch);
          Exit;
        end;
      end else if precalc[1] = '+' then
      begin
        {plus has no effect}
      end else
      begin
        DoPreCalc := False;
        Exit;
      end;
      Delete(PreCalc, 1, 1);
    end;
  end;

  procedure DisposeList;
    { Dispose the items }
  var
    i: Integer;
    p: PSmallCalculation;
  begin
    for i := 0 to Items.Count - 1 do begin
      p := items.GetItem(i);
      if p^.TType = 0 then
        DestroyCajVariant(p^.CajVariant);
      Dispose(p);
    end;
    Items.Destroy;
  end;
  function ParseString: string;
    { Parse a string }
  var
    temp3: string;
  begin
    temp3 := '';
    while (Parser^.CurrTokenId = CSTI_String) or
      (Parser^.CurrTokenId = CSTI_Char) do
    begin
      if Parser^.CurrTokenId = CSTI_String
        then
      begin
        temp3 := temp3 + PString(GetToken(Parser));
        SkipWhiteSpace(Parser);
        if Parser^.CurrTokenId = CSTI_String then
          temp3 := temp3 + #39;
      end {if}
      else
      begin
        temp3 := temp3 + ChrToStr(GetToken(Parser));
        SkipWhiteSpace(parser);
      end; {else if}
    end; {while}
    ParseString := temp3;
  end;
  procedure Calculate;
    { Calculate the full expression }
  var
    l: PSmallCalculation;
    i: LongInt;
  begin
    i := 0;
    while i < (items.count - 1) div 2 do begin
      l := PSmallCalculation(items.GetItem(i * 2 + 1));
      if (l^.TType >= 2) and (l^.TType <= 8) then
      begin
        case l^.TType of
          2: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtMul)
            then
              RunError(ETypeMismatch);
          3: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtDiv)
            then
              RunError(ETypeMismatch);
          4: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtIntDiv)
            then
              RunError(ETypeMismatch);
          5: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtIntMod)
            then
              RunError(ETypeMismatch);
          6: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtAnd)
            then
              RunError(ETypeMismatch);
          7: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtShr)
            then
              RunError(ETypeMismatch);
          8: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtShl)
            then
              RunError(ETypeMismatch);
        end;
        if ErrorCode <> 0 then
          Exit;
        l := PSmallCalculation(items.GetItem(i * 2 + 2));
        DestroycajVariant(l^.CajVariant);
        Dispose(l);
        Items.Remove(l);
        l := PSmallCalculation(items.GetItem(i * 2 + 1));
        Dispose(l);
        Items.Remove(l);
      end else Inc(i);
    end;

    i := 0;
    while i < (items.count - 1) div 2 do begin
      l := PSmallCalculation(items
        .GetItem(i * 2 + 1));
      if (l^.TType >= 9) and (l^.TType <= 12) then
      begin
        case l^.TType of
          9: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtPlus)
            then
              RunError(ETypeMismatch);
          10: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtMinus)
            then
              RunError(ETypeMismatch);
          11: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtOr)
            then
              RunError(ETypeMismatch);
          12: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtXor)
            then
              RunError(ETypeMismatch);
        end;
        if ErrorCode <> 0 then
          Exit;
        l := PSmallCalculation(items.GetItem(i * 2 + 2));
        DestroycajVariant(l^.CajVariant);
        Dispose(l);
        Items.Remove(l);
        l := PSmallCalculation(items
          .GetItem(i * 2 + 1));
        Dispose(l);
        Items.Remove(l);
      end else Inc(i);
    end;
    i := 0;
    while i < (items.count - 1) div 2 do begin
      l := PSmallCalculation(items
        .GetItem(i * 2 + 1));
      if (l^.TType >= 13) and (l^.TType <= 18) then
      begin
        case l^.TType of
          13: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtEqual)
            then
              RunError(ETypeMismatch);
          14: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtGreater)
            then
              RunError(ETypeMismatch);
          15: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtLess)
            then
              RunError(ETypeMismatch);
          16: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtNotEqual)
            then
              RunError(ETypeMismatch);
          17: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtLessEqual)
            then
              RunError(ETypeMismatch);
          18: if not Perform(Types, PSmallCalculation(items.GetItem(i * 2))^.
              CajVariant, PSmallCalculation(items.GetItem(i * 2 + 2))
              ^.CajVariant, PtGreaterEqual)
            then
              RunError(ETypeMismatch);
        end;
        if ErrorCode <> 0 then
          Exit;
        l := PSmallCalculation(items.GetItem(i * 2 + 2));
        DestroycajVariant(l^.CajVariant);
        Dispose(l);
        Items.Remove(l);
        l := PSmallCalculation(items.GetItem(i * 2 + 1));
        Dispose(l);
        Items.Remove(l);
      end else Inc(i);
    end;
  end;
begin
{$IFDEF CLASSLIST}
  Items := TList.Create;
{$ELSE}
  Items.Create;
{$ENDIF}
  Calc := False;
  Res := GetVarLink(res);
  while True do begin
    if Parser^.CurrTokenId = StopOn then
      Break;
    case Parser^.CurrTokenId of
      CSTII_Else,
        CSTII_To,
        CSTII_DownTo,
        CSTII_do,
        CSTII_until,
        CSTI_Semicolon,
        CSTII_End,
        CSTI_Comma,
        CSTI_CloseRound:
        begin
          Break;
        end; {Csti_Else...}
      CSTI_EOF:
        begin
          RunError(EUnexpectedEndOfFile);
          DisposeList;
          Exit;
        end; {CSTI_Eof}
      CSTI_SyntaxError,
        CSTI_CommentEOFError,
        CSTI_CharError:
        begin
          RunError(ESyntaxError);
          DisposeList;
          Exit;
        end; {Csti_SyntaxError...}
      CSTI_StringError:
        begin
          RunError(EStringError);
          DisposeList;
          Exit;
        end; {csti_Stringerror}
    end; {case}
    if (Items.Count and $1) = 0 then
    begin
      PreCalc := '';
      while (Parser^.CurrTokenId = CSTI_Minus) or
        (Parser^.CurrTokenId = CSTII_Not) or
        (Parser^.CurrTokenId = CSTI_Plus)
        do begin
        if (Parser^.CurrTokenId = CSTI_Minus) then
          PreCalc := PreCalc + '-';
        if (Parser^.CurrTokenId = CSTII_Not) then
          PreCalc := PreCalc + '|';
        if (Parser^.CurrTokenId = CSTI_Plus) then
          PreCalc := PreCalc + '+';
        SkipWhiteSpace(Parser);
      end; {While}

      New(Work);
      case Parser^.CurrTokenId of
        CSTI_OpenRound:
          begin
            SkipWhiteSpace(Parser);
            Work^.CajVariant := CreateCajVariant(TM_Add(Types, '', CSV_Var, nil));
            Work^.CajVariant^.Cv_Var := nil;
            Work^.TType := 0;
            if not Calc(vars, Work^.CajVariant, CSTI_CloseRound, OnlyConst) then
            begin
              DestroyCajVariant(Work^.CajVariant);
              Dispose(Work);
              DisposeList;
              Exit;
            end; {if}
            if not DoPreCalc then
            begin
              DestroyCajVariant(Work^.CajVariant);
              Dispose(Work);
              DisposeList;
              Exit;
            end; {if}
            SkipWhiteSpace(Parser);
            Items.Add(Work);
          end; {CSTI_OpenRound}
        CSTI_Identifier:
          begin
            if Assigned(vars) and (Vm_Find(Vars, FastUppercase(GetToken(Parser)))
              <> -1) then
            begin
              Temp4 := GetVarLink(Vm_Get(Vars, Vm_Find(Vars,
                FastUppercase(GetToken(Parser)))));
              if ((Temp4^.Flags and $1) = 0) and (OnlyConst) then
              begin
                RunError(EConstantExpected);
                Dispose(Work);
                DisposeList;
                Exit;
              end;
              if not CalcSub(Vars, Temp4) then
              begin
                Dispose(Work);
                DisposeList;
                Exit;
              end;
              SkipWhiteSpace(Parser);
              Work^.CajVariant := CreateCajVariant(Temp4^.VType);
              Work^.TType := 0;
              if not PerForm(Types, Work^.CajVariant, Temp4, ptSet) then
              begin
                DestroyCajVariant(Work^.CajVariant);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
              if not DoPreCalc then
              begin
                DestroyCajVariant(Work^.CajVariant);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
            end {if}
            else if Vm_Find(Variables, FastUppercase(GetToken(Parser))) <> -1
              then
            begin
              Temp4 := GetVarLink(Vm_Get(Variables,
                Vm_Find(Variables, FastUppercase(GetToken(Parser)))));
              if ((Temp4^.Flags and $1) = 0) and (OnlyConst) then
              begin
                RunError(EConstantExpected);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
              if not CalcSub(Vars, Temp4) then
              begin
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
              SkipWhiteSpace(Parser);
              Work^.CajVariant := CreateCajVariant(Temp4^.VType);
              Work^.TType := 0;
              if not PerForm(Types, Work^.CajVariant, Temp4, ptSet) then
              begin
                RunError(EErrorInExpression);
                DestroyCajVariant(Work^.CajVariant);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
              if not DoPreCalc then
              begin
                DestroyCajVariant(Work^.CajVariant);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
            end {if}
            else if PM_Find(Procedures, FastUpperCase(GetToken(Parser))) <> -1 then
            begin
              if OnlyConst then
              begin
                RunError(EConstantExpected);
                Dispose(Work);
                DisposeList;
                Exit;
              end;
              Temp4 := DoProc(vars, False);
              if Temp4 = nil then
              begin
                RunError(EConstantExpected);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
              Work^.CajVariant := CreateCajVariant(Temp4^.VType);
              Work^.TType := 0;
              PerForm(Types, Work^.CajVariant, Temp4, ptSet);
              if not DoPreCalc then
              begin
                DestroyCajVariant(Work^.CajVariant);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
            end {else if}
            else if PM_Find(InternalProcedures, FastUpperCase(GetToken(Parser)))
              <> -1 then
            begin
              if OnlyConst then begin
                RunError(EConstantExpected);
                Dispose(Work);
                DisposeList;
                Exit;
              end;
              Temp4 := DoProc(vars, True);
              if ErrorCode <> ENoError then
              begin
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
              Work^.CajVariant := CreateCajVariant(Temp4^.VType);
              Work^.TType := 0;
              PerForm(Types, Work^.CajVariant, Temp4, ptSet);
              if not DoPreCalc then
              begin
                DestroyCajVariant(Work^.CajVariant);
                Dispose(Work);
                DisposeList;
                Exit;
              end; {if}
            end {else if}
            else
            begin
              RunError(EUnknownIdentifier);
              Dispose(Work);
              DisposeList;
              Exit;
            end; {else else if}
            Items.Add(Work);
          end; {CSTI_Identifier}
        CSTI_Integer:
          begin
            if ((Res^.VType^.atypeid >= csv_SByte) and (Res^.VType^.atypeid <= Csv_SInt32)) or
              ((res^.VType^.atypeid >= csv_Real) and (res^.vtype^.atypeid <= csv_comp)) then
              Work^.CajVariant := CreateCajVariant(res^.VType)
            else
              Work^.CajVariant := CreateCajVariant(TM_ADD(Types, '', csv_SInt32, nil));
            Work^.TType := 0;
            if IsRealType(work^.CajVariant) then
              SetReal(Work^.CajVariant, StrToInt(GetToken(Parser)))
            else
              SetInteger(Work^.CajVariant, StrToInt(GetToken(Parser)));
            if not DoPreCalc then
            begin
              DestroyCajVariant(Work^.CajVariant);
              Dispose(Work);
              DisposeList;
              Exit;
            end; {if}
            SkipWhiteSpace(Parser);
            Items.Add(Work);
          end; {CSTI_Integer}
        CSTI_Real:
          begin
            if (Res^.VType^.atypeid >= CSV_Real) and (Res^.VType^.atypeid <= CSV_Comp) then
              Work^.CajVariant := CreateCajVariant(res^.VType)
            else
              Work^.CajVariant := CreateCajVariant(TM_Add(Types, '', CSV_Extended, nil));
            Work^.TType := 0;
            SetReal(Work^.CajVariant, StrToReal(GetToken(Parser)));
            if not DoPreCalc then
            begin
              DestroyCajVariant(Work^.CajVariant);
              Dispose(Work);
              DisposeList;
              Exit;
            end;
            SkipWhiteSpace(Parser);
            Items.Add(Work);
          end; {CSTI_Real}
        CSTI_String, CSTI_Char:
          begin

            Work^.CajVariant := CreateCajVariant(TM_Add(Types, '', CSV_STRING, nil));
            Work^.TType := 0;
            Work^.CajVariant^.CV_Str := ParseString;
            if not DoPreCalc then
            begin
              DestroyCajVariant(Work^.CajVariant);
              Dispose(Work);
              DisposeList;
              Exit;
            end; {if}
            Items.Add(Work);
          end; {CSTI_String}
        CSTI_HexInt:
          begin
            Work^.TType := 0;
            if (Res^.VType^.atypeid >= csv_SByte) and (Res^.VType^.aTypeId <= Csv_SInt32) then
              Work^.CajVariant := CreateCajVariant(res^.VType)
            else
              Work^.CajVariant := CreateCajVariant(TM_Add(Types, '', CSV_SInt32, nil));
            SetInteger(Work^.CajVariant, StrToInt(GetToken(Parser)));
            if not DoPreCalc then
            begin
              DestroyCajVariant(Work^.CajVariant);
              Dispose(Work);
              DisposeList;
              Exit;
            end; {if}
            SkipWhiteSpace(Parser);
            Items.Add(Work);
          end; {CSTI_HexInt}
      else
        begin
          RunError(EErrorInExpression);
          Dispose(Work);
          DisposeList;
          Exit;
        end;
      end; {case}
    end {if}
    else
    begin
      New(Work);
      case Parser^.CurrTokenId of
        CSTI_Equal: Work^.TType := 13;
        CSTI_NotEqual: Work^.TType := 16;
        CSTI_Greater: Work^.TType := 14;
        CSTI_GreaterEqual: Work^.TType := 18;
        CSTI_Less: Work^.TType := 15;
        CSTI_LessEqual: Work^.TType := 17;
        CSTI_Plus: Work^.TType := 9;
        CSTI_Minus: Work^.TType := 10;
        CSTI_Divide:
          begin
            Work^.TType := 3;
            if Res^.VType^.atypeid = CSV_VAR then
              SetType(Res, TM_Add(Types, '', CSV_Extended, nil));
          end;
        CSTI_Multiply: Work^.TType := 2;
        CSTII_and: Work^.TType := 6;
        CSTII_div: Work^.TType := 4;
        CSTII_mod: Work^.TType := 5;
        CSTII_or: Work^.TType := 11;
        CSTII_shl: Work^.TType := 8;
        CSTII_shr: Work^.TType := 7;
        CSTII_xor: Work^.TType := 12;
      else
        begin
          RunError(EErrorInExpression);
          Dispose(Work);
          DisposeList;
          Exit;
        end; {else case}
      end; {case}
      Items.Add(Work);
      SkipWhiteSpace(parser);
    end; {else if}
  end; {while}
  Calculate;
  if ErrorCode = 0 then
  begin
    if Items.Count <> 1 then
    begin
      RunError(EErrorInExpression);
      Calc := False;
    end else
    begin
      Work := Items.GetItem(0);
      if Perform(Types, Res, Work^.CajVariant, PtSet) then
        Calc := True
      else RunError(ETypeMismatch);
    end; {if}
  end; {if}
  DisposeList;
end; {Calc}

function TIfPasScript.RunScriptProc(const Name: string; Parameters:
  PVariableManager): PIfVariant;
var
  ProcCall: LongInt;
  ProcDef: string;
  w: PIfVariant;
  i: LongInt;
  function IRem(S: string): string;
  {Remove the !}
  begin
    Delete(s, 1, 1);
    IRem := s;
  end; {irem}
begin
  RunError(0);
  RunScriptProc := nil;
  if MainOffset = -1 then
  begin
    Parser^.CurrTokenPos := -1;
    RunError(EBeginExpected);
    Exit;
  end; {if}
  if PM_Find(InternalProcedures, FastUpperCase(Name)) = -1 then
  begin
    Parser^.CurrTokenPos := 0;
    RunError(EUnknownIdentifier);
    Exit;
  end; {if}
  ProcCall := LongInt(PM_Get(InternalProcedures, PM_Find(InternalProcedures,
    FastUpperCase(Name))));
  ProcDef := PM_GetSpec(InternalProcedures, PM_Find(InternalProcedures,
    FastUpperCase(Name)));
  if IntProcDefParam(ProcDef, -1) <> VM_Count(Parameters) then
  begin
    Parser^.CurrTokenPos := 0;
    RunError(EParameterError);
    Exit;
  end;
  for i := 1 to IntProcDefParam(ProcDef, -1) do begin
    if Pos('!', IntProcDefName(ProcDef, I)) = 1 then
    begin
      w := GetVarLink(VM_Get(Parameters, i - 1));
      if (Longint(w^.VType) <> IntProcDefParam(ProcDef, I)) or ((Longint(W^.Flags) and $1) <> 0) then
      begin
        Parser^.CurrTokenPos := -I;
        RunError(EParameterError);
        Exit;
      end; {if}
      VM_SetName(Parameters, I - 1, IRem(IntProcDefName(ProcDef, I)));
    end {if} else
    begin
      w := GetVarLink(VM_Get(Parameters, i - 1));
      if IntProcDefParam(ProcDef, i) <> Longint(w^.VType) then
      begin
        Parser^.CurrTokenPos := -I;
        RunError(EParameterError);
        Exit;
      end; {if}
      VM_SetName(Parameters, I - 1, IntProcDefName(ProcDef, I));
    end; {else if}
  end; {for}
  if IntProcDefParam(ProcDef, 0) <> 0 then
  begin
    w := CreateCajVariant(Pointer(IntProcDefParam(ProcDef, 0)));
    VM_Add(Parameters, CreateCajVariant(TM_Add(Types, '', CSV_Var, nil)), 'RESULT')^.Cv_Var := w;
  end {if}
  else w := nil;
  Parser^.CurrTokenPos := ProcCall;
  ParseToken(Parser);
  if Parser^.CurrTokenId = CSTII_Var then
  begin
    if not ProcessVars(Parameters) then
    begin
      DestroyCajVariant(w);
      Exit;
    end; {if}
  end; {if}
  if not RunBegin(Parameters, False) then
  begin
    if FErrorCode <> EExitCommand then
    begin
      DestroycajVariant(w);
      Exit;
    end else
      FErrorCode := 0;
  end; {if}
  ParseToken(Parser);
  RunScriptProc := w;

end;

function TIfPasScript.DoProc(Vars: PVariableManager; Internal: Boolean): PIfVariant;
{Call an internal/external Procedure}
var
  ProcCall: TRegisteredProc;
  ProcCall2: LongInt;
  ProcDef: string;
  w: PIfVariant;
  i: LongInt;
  Params: PVariableManager;
  function IRem(S: string): string;
  {Remove the !}
  begin
    Delete(s, 1, 1);
    IRem := s;
  end; {irem}
begin
  DoProc := nil;
  if Internal then
  begin
    ProcCall2 := LongInt(PM_Get(InternalProcedures, PM_Find(InternalProcedures,
      FastUpperCase(GetToken(Parser)))));
    ProcDef := PM_GetSpec(InternalProcedures, PM_Find(InternalProcedures,
      FastUpperCase(GetToken(Parser))));
{$IFDEF FPC}
    ProcCall := TRegisteredProc(nil);
{$ELSE}
    ProcCall := nil;
{$ENDIF}
  end else
  begin
{$IFDEF FPC}
    ProcCall := TRegisteredProc(PM_Get(Procedures, PM_Find(Procedures,
      FastUpperCase(GetToken(Parser)))));
{$ELSE}
    @ProcCall := PM_Get(Procedures, PM_Find(Procedures,
      FastUpperCase(GetToken(Parser))));
{$ENDIF}
    ProcCall2 := 0;
    ProcDef := PM_GetSpec(Procedures, PM_Find(Procedures,
      FastUpperCase(GetToken(Parser))));
  end;
  Params := VM_Create(nil);
  SkipWhiteSpace(Parser);
  if (IntProcDefParam(ProcDef, -1) <> 0) and (Parser^.CurrTokenId <>
    CSTI_OpenRound) then
  begin
    RunError(ERoundOpenExpected);
    VM_Destroy(params);
    Exit;
  end; {if}
  if (IntProcDefParam(ProcDef, -1) = 0) and (Parser^.CurrTokenId =
    CSTI_OpenRound) then
  begin
    SkipWhiteSpace(Parser);
    if Parser^.CurrTokenId = CSTI_CloseRound then
    begin
      SkipWhiteSpace(Parser);
    end else
    begin
      RunError(ECloseRoundExpected);
      VM_Destroy(params);
      Exit;
    end;
  end; {if}
  if Parser^.CurrTokenId = CSTI_OpenRound then
  begin
    for i := 1 to IntProcDefParam(ProcDef, -1) do begin
      SkipWhiteSpace(Parser);
      if Pos('!', intProcDefName(ProcDef, i)) = 1 then
      begin
        {Expect a variable}
        if Assigned(Vars) and (VM_Find(Vars, FastUppercase(GetToken(Parser))) <>
          -1) then
          w := GetVarLink(VM_Get(Vars, VM_Find(Vars,
            FastUppercase(GetToken(Parser)))))
        else if VM_Find(Variables, FastUppercase(GetToken(Parser))) <> -1 then
          w := GetVarLink(VM_Get(Variables, VM_Find(Variables,
            FastUppercase(GetToken(Parser)))))
        else
        begin
          RunError(EVariableExpected);
          VM_Destroy(params);
          Exit;
        end; {else else if}
        if (w^.Flags and $1) <> 0 then
        begin
          RunError(EVariableExpected);
          VM_Destroy(params);
          Exit;
        end; {if}
        if not CalcSub(Vars, w) then
        begin
          VM_Destroy(params);
          Exit;
        end;
        if (Longint(w^.VType) <> IntProcDefParam(ProcDef, I)) and not ((PTypeRec(IntProcDefParam(ProcDef, I))^.ext = nil) and (PTypeRec(IntProcDefParam(ProcDef, I))^.atypeid = CSV_Array)) then
        begin
          RunError(ETypeMismatch);
          VM_Destroy(params);
          Exit;
        end;
        VM_Add(Params, CreateCajVariant(TM_Add(Types, '', CSV_Var, nil)),
          FastUppercase(IRem(IntProcDefName(ProcDef, i))))^.Cv_var := w;
        SkipWhiteSpace(Parser);
      end {if}
      else
      begin
        w := VM_Add(Params, CreateCajVariant(Pointer(IntProcDefParam(ProcDef, i))),
          IntProcDefName(ProcDef, i));
        if not Calc(vars, w, CSTI_CloseRound, False) then
        begin
          VM_Destroy(params);
          Exit;
        end; {if}
      end; {else if}
      if i = IntProcDefParam(ProcDef, -1) then
      begin
        if parser^.CurrTokenId <> CSTI_CloseRound then
        begin
          RunError(ERoundCloseExpected);
          VM_Destroy(params);
          Exit;
        end; {if}
      end {if}
      else
      begin
        if parser^.CurrTokenId <> CSTI_Comma then
        begin
          RunError(ECommaExpected);
          VM_Destroy(params);
          Exit;
        end; {if}
      end; {else if}
    end; {for}
    SkipWhiteSpace(Parser);
  end; {if}
  {Now we have all the parameters}
  if Internal then
  begin
    if IntProcDefParam(ProcDef, 0) <> 0 then
    begin
      w := CreateCajVariant(Pointer(IntProcDefParam(ProcDef, 0)));
      VM_Add(Params, CreateCajVariant(TM_Add(Types, '', CSV_Var, nil)), 'RESULT')^.Cv_Var := w;
    end {if}
    else w := nil;
    i := Parser^.CurrTokenPos;
    Parser^.CurrTokenPos := ProcCall2;
    ParseToken(Parser);
    if Parser^.CurrTokenId = CSTII_Var then
    begin
      if not ProcessVars(Params) then
      begin
        DestroyCajVariant(w);
        Exit;
      end; {if}
    end; {if}
    if not RunBegin(Params, False) then
    begin
      if FErrorCode = EExitcommand then
        FErrorCode := 0
      else begin
        DestroycajVariant(w);
        Exit;
      end;
    end; {if}
    Parser^.CurrTokenPos := I;
    ParseToken(Parser);
    DoProc := w;
    VM_Destroy(params);
  end {if}
  else
  begin
    if IntProcDefParam(ProcDef, 0) <> 0 then
      w := CreateCajVariant(Pointer(IntProcDefParam(ProcDef, 0)))
    else
      w := nil;
    RunError(ProcCall(fId, IntProcDefName(ProcDef, 0), Params, w));
    if ErrorCode <> ENoError then
    begin
      VM_Destroy(params);
      DestroyCajVariant(w);
      Exit;
    end; {if}
    VM_Destroy(params);
    DoProc := w;
  end; {if}
end; {DoExternalProc}

//-------------------------------------------------------------------
// Procedure: RunBegin
//   Purpose: Steps through the script, parsing the tokens
//-------------------------------------------------------------------

function TIfPasScript.RunBegin(Vars: PVariableManager; Skip: Boolean): Boolean;
      { Run the Script, this is the main part of the script engine }
var
  StopOnSemicolon: Boolean;
  c, c2: PIfVariant;
  IPos, IStart, II, IEnd: LongInt;
  B: Boolean;
  lBreak: boolean;
  lRepeat: Boolean;

begin
  lRepeat := Parser^.CurrTokenId = CSTII_repeat;
  RunBegin := False;
  if Skip then
  begin
    if Parser^.CurrTokenId = CSTII_Begin then
    begin
      SkipWhiteSpace(Parser); // Find non-whitespace, comment token
      IPos := 1;
      while True do
      begin
        if Parser^.CurrTokenId = CSTI_EOF then
        begin
          RunError(EUnexpectedEndOfFile);
          Exit;
        end;
        if Parser^.CurrTokenId = CSTI_SyntaxError then
        begin
          RunError(ESyntaxError);
          Exit;
        end;
        if Parser^.CurrTokenId = CSTI_CommentEOFError then
        begin
          RunError(ESyntaxError);
          Exit;
        end;
        if Parser^.CurrTokenId = CSTI_CharError then
        begin
          RunError(ESyntaxError);
          Exit;
        end;
        if Parser^.CurrTokenId = CSTI_StringError then
        begin
          RunError(EStringError);
          Exit;
        end;
        if Parser^.CurrTokenId = CSTII_Case then
          Inc(IPos);
        if Parser^.CurrTokenId = CSTII_Begin then
          Inc(IPos);
        if Parser^.CurrTokenId = CSTII_Repeat then
          Inc(IPos);
        if (Parser^.CurrTokenId = CSTII_End) or (Parser^.CurrTokenID = CSTII_Until) then
        begin
          Dec(IPos);
          if IPos = 0 then
          begin
            SkipWhiteSpace(Parser);
            Break;
          end;
        end;
        SkipWhiteSpace(Parser);
      end; {While}
      SkipWhiteSpace(Parser); {Skip end}
    end else
    begin
      IPos := 1;
      while True do begin
        if Parser^.CurrTokenId = CSTI_EOF then
        begin RunError(EUnexpectedEndOfFile); Exit; end;
        if Parser^.CurrTokenId = CSTI_SyntaxError then
        begin RunError(ESyntaxError); Exit; end;
        if Parser^.CurrTokenId = CSTI_CommentEOFError then
        begin RunError(ESyntaxError); Exit; end;
        if Parser^.CurrTokenId = CSTI_CharError then
        begin RunError(ESyntaxError); Exit; end;
        if Parser^.CurrTokenId = CSTI_StringError then
        begin RunError(EStringError); Exit; end;
        if Parser^.CurrTokenId = CSTI_SemiColon then
          Break;
        if Parser^.CurrTokenId = CSTII_Begin then
        begin
          if not RunBegin(Vars, True) then
            Exit;
        end;
        if Parser^.CurrTokenId = CSTII_Else then
        begin
          Dec(IPos);
          if Ipos <= 0 then
            Break;
        end;
        if Parser^.CurrTokenId = CSTII_If then
          Inc(IPos);
        if Parser^.CurrTokenId = CSTII_Repeat then
          Inc(IPos);
        if Parser^.CurrTokenId = CSTII_End then
          Break;
        if Parser^.CurrTokenId = CSTII_Until then
          Break;
        SkipWhiteSpace(Parser);
      end; {While}
    end; {If}
    RunBegin := True;
    Exit;
  end; {If}

  if (Parser^.CurrTokenId = CSTII_Begin) or (Parser^.CurrTokenID = CSTII_Repeat) then
  begin
    StopOnSemicolon := False;
    SkipWhiteSpace(Parser); {skip begin}
  end
  else
    StopOnSemicolon := True;

  while True do begin

    case Parser^.CurrTokenId of
      CSTI_EOF: begin RunError(EUnexpectedEndOfFile); Exit; end;
      CSTI_SyntaxError: begin RunError(ESyntaxError); Exit; end;
      CSTI_CommentEOFError: begin RunError(ESyntaxError); Exit; end;
      CSTI_CharError: begin RunError(ESyntaxError); Exit; end;
      CSTI_StringError: begin RunError(EStringError); Exit; end;
      CSTII_Else: begin
          if StopOnSemicolon
            then RunBegin := True
          else begin
            RunError(EErrorInStatement);
            RunBegin := False;
            Exit;
          end;
        end;
      CSTII_End:
        begin
          if not lRepeat then
          begin
            RunBegin := True;
            SkipWhiteSpace(Parser);
            Exit;
          end else
          begin
             RunError(EUntilExpected);
             RunBegin := False;
             exit;
          end;
        end; {CSTII_End}
      CSTII_Until:
        begin
          if lRepeat then
          begin
            RunBegin := True;
            Exit;
          end else
          begin
             RunError(EEndExpected);
             RunBegin := False;
             exit;
          end;
        end; {CSTII_Until}
//-------------------------------------------------------
// Exit command - aborts the script
//-------------------------------------------------------
      CSTII_Exit:
        begin
          RunBegin := False;
          RunError(EExitCommand);
          exit;
        end; { CSTII_Exit}
//-------------------------------------------------------
// Break command - breaks out of loop
//-------------------------------------------------------
      CSTII_Break:
        begin
          RunBegin := true;
//             SkipWhiteSpace(Parser);
          exit;
        end; { CSTII_Break }

//-------------------------------------------------------
// Semicolon is handled specially
//-------------------------------------------------------
      CSTI_Semicolon:
        begin
          if StopOnSemicolon then
          begin
            RunBegin := True;
            Exit;
          end;
          SkipWhiteSpace(Parser);
        end; {CSTI_SemiColon}
//-------------------------------------------------------
// Process an IF statement
//-------------------------------------------------------
      CSTII_If:
        begin
          if Assigned(OnRunLine) then
          begin
            RunError(OnRunLine(fId, Self, Parser^.CurrTokenPos));
            if ErrorCode <> ENoError then
              Exit;
          end;
          SkipWhiteSpace(Parser);
          c := CreateCajVariant(TM_Add(Types, '', CSV_Bool, nil));
          if not Calc(vars, c, CSTII_Then, False) then
          begin
            DestroyCajVariant(c);
            Exit;
          end; {if}
          if Parser^.CurrTokenId <> CSTII_then
            then
          begin
            RunError(EThenExpected);
            DestroyCajVariant(c);
            Exit;
          end;
          SkipWhiteSpace(Parser); {skip THEN}
          if c^.Cv_Bool then
          begin
            DestroyCajVariant(c);
            if not RunBegin(Vars, False) then
            begin
              Exit;
            end; {if}
            if Parser^.CurrTokenId = CSTII_Else then
            begin
              if Assigned(OnRunLine) then
              begin
                RunError(OnRunLine(fId, Self, Parser^.CurrTokenPos));
                if ErrorCode <> ENoError then
                  Exit;
              end;
              SkipWhiteSpace(Parser);
              if not RunBegin(Vars, True) then
              begin
                Exit;
              end; {if}
            end; {if}
          end {if}
          else
          begin
            DestroyCajVariant(c);
            if not RunBegin(Vars, True) then
            begin
              Exit;
            end; {if}
            if Parser^.CurrTokenId = CSTII_Else then
            begin
              SkipWhiteSpace(Parser);
              if Assigned(OnRunLine) then
              begin
                RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
                if ErrorCode <> ENoError then
                  Exit;
              end;
              if not RunBegin(Vars, False) then
              begin
                Exit;
              end; {if}
            end; {if}
          end; {if}
        end; {CSTII_If}

//--------------------------------------------------------------------
// Process the WHILE DO loop ***************************************--
//--------------------------------------------------------------------
      CSTII_While:
        begin
          lBreak := false;

          SkipWhiteSpace(Parser); // Find Next token

          C := CreateCajVariant(TM_Add(Types, '', CSV_Bool, nil)); // Create a boolean variable
          Ipos := Parser^.CurrTokenPos; // Save position of variable

          // Test the expression up to the DO command
          //-------------------------------------------
          if not Calc(Vars, c, CSTII_Do, False) then
          begin
            DestroyCajVariant(c);
            Exit;
          end; {if}

          // If not a DO command, this is an error
          //-----------------------------------------
          if Parser^.CurrTokenID <> CSTII_Do then
          begin
            RunError(EDoExpected);
            DestroyCajVariant(c);
            Exit;
          end;

          SkipWhiteSpace(Parser);

          // Save the DO block starting position
          //--------------------------------------
          IStart := Parser^.CurrTokenPos;

          //-------------------------------
          // Start the loop processing
          //-------------------------------
          while C^.Cv_Bool and (not lBreak) do
          begin
             // See if any command to run
             // ---------------------------
            if Assigned(OnRunLine) then
            begin
              RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
              if ErrorCode <> ENoError then Exit;
            end;

             // Call routine to process the code
             //----------------------------------
            if not RunBegin(Vars, False) then
            begin
              DestroyCajVariant(c);
              Exit;
            end;

            // If a break were returned
            //--------------------------
            if Parser^.CurrTokenId = CSTII_Break then
            begin
              lBreak := true;
            end;

            // Go back to the variable test position
            //----------------------------------------
            Parser^.CurrTokenPos := IPos;
            ParseToken(Parser);

            // Test the condition again, if false, exit
            //-----------------------------------------
            if not Calc(Vars, c, CSTII_Do, False) then
            begin
              DestroyCajVariant(c);
              Exit;
            end;

            // Reset to the DO starting block
            //---------------------------------
            Parser^.CurrTokenPos := IStart;
            ParseToken(Parser);
          end;


          DestroyCajVariant(c);
          if not RunBegin(Vars, True) then Exit;
        end;
//--------------------------------------------------------------------
// Process the REPEAT UNTIL loop ***********************************--
//--------------------------------------------------------------------
      CSTII_Repeat:
        begin
          lBreak := false;

          C := CreateCajVariant(TM_Add(Types, '', CSV_Bool, nil)); // Create a boolean variable

          IStart := Parser^.CurrTokenPos;

          //-------------------------------
          // Start the loop processing
          //-------------------------------
          repeat
            Parser^.CurrTokenPos := IStart;
            ParseToken(Parser);
             // See if any command to run
             // ---------------------------
            if Assigned(OnRunLine) then
            begin
              RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
              if ErrorCode <> ENoError then Exit;
            end;

             // Call routine to process the code
             //----------------------------------
            if not RunBegin(Vars, False) then
            begin
              DestroyCajVariant(c);
              Exit;
            end;

            // If a break were returned
            //--------------------------
            if Parser^.CurrTokenId = CSTII_Break then
            begin
              lBreak := true;
              Break;
            end;

            if Parser^.CurrTokenID <> CSTII_Until then
            begin
              RunError(EUntilExpected);
              DestroyCajVariant(C);
              Exit;
            end;

            SkipWhiteSpace(Parser);

            // Test the condition again, if false, exit
            //-----------------------------------------
            if not Calc(Vars, c, CSTI_SemiColon, False) then
            begin
              DestroyCajVariant(c);
              Exit;
            end;
          until C^.Cv_Bool;

          DestroyCajVariant(c);

          if lBreak then
          begin
            Parser^.CurrTokenPos := IStart;
            ParseToken(Parser);
            if not RunBegin(Vars, True) then Exit;
          end;
        end;

//-------------------------------------------------------
// FOR LOOP is handled here
//-------------------------------------------------------
      CSTII_For:
        begin
          // Found the FOR keyword
          // ----------------------
          SkipWhiteSpace(Parser); // Find the next token

          // It should be an identifier
          //------------------------------
          if Parser^.CurrTokenId <> CSTI_Identifier then
          begin
            RunError(EIdentifierExpected);
            Exit;
          end; {if}

          // The variable must exist
          //------------------------------
          if Assigned(Vars) and
            (VM_Find(Vars, FastUppercase(GetToken(Parser))) <> -1)
            then C := GetVarLink(VM_Get(Vars, VM_Find(Vars,
              FastUppercase(GetToken(Parser)))))
          else
            if VM_Find(Variables, FastUppercase(GetToken(Parser))) <> -1
              then c := GetVarLink(VM_Get(Variables, VM_Find(Variables,
                FastUppercase(GetToken(Parser)))))
            else
            begin
              RunError(EUnknownIdentifier);
              Exit;
            end; {if}

          // It cannot be a constant
          //------------------------------
          if (c^.Flags and $1) <> 0 then
          begin
            RunError(EVariableExpected);
            Exit;
          end; {if}

          // And it must be an integer
          //------------------------------
          if not IsIntegerType(c) then
          begin
            RunError(eIntegerExpected);
          end; {if}

          SkipWhiteSpace(Parser); // Find the next token

          // Expecting an assignment statement
          //----------------------------------------
          if Parser^.CurrTokenId <> CSTI_Assignment then
          begin
            RunError(EAssignmentExpected);
            Exit;
          end; {if}

          SkipWhiteSpace(Parser); // Find the next token

          // Calculate expression of token from current position
          // to the TO keyword
          //--------------------------------------------------
          if not Calc(Vars, c, CSTII_To, False)
            then Exit;

          // Get the result of the calculation
          //---------------------------------------
          IStart := GetInteger(c);


          if (Parser^.CurrTokenId <> CSTII_To) and
            (Parser^.CurrTokenId <> CSTII_DownTo) then
          begin
            RunError(EToExpected);
            Exit;
          end; {if}

          // See if we are going up or down
          //-----------------------------------
          B := (Parser^.CurrTokenId = CSTII_DownTo);

          SkipWhiteSpace(Parser); // Find the next token

          if not Calc(Vars, c, CSTII_Do, False)
            then Exit;

          // Get the result of the calculation
          //---------------------------------------
          IEnd := GetInteger(c);


          if Parser^.CurrTokenId <> CSTII_Do then
          begin
            RunError(EDoExpected);
            Exit;
          end; {if}


          SkipWhiteSpace(Parser); // Find the next token

          lBreak := false; // Assume the loop will complete
          IPos := Parser^.CurrTokenPos;

          if B then
          begin
            c^.Flags := c^.Flags or $1;

            // Start the loop
            //------------------------------
            for II := IStart downto IEnd do
            begin
              if Assigned(OnRunLine) then
              begin
                RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
                if ErrorCode <> ENoError then Exit;
              end;

              // Make the loop variable visible to program
              //--------------------------------------------
              SetInteger(C, II);
              if not RunBegin(Vars, False) then
              begin
                c^.Flags := c^.Flags and not $1;
                Exit;
              end;

              // If a break were returned
              //--------------------------
              if Parser^.CurrTokenId = CSTII_Break
                then lBreak := true;

              Parser^.CurrTokenPos := IPos;
              ParseToken(Parser);

              if lBreak then break;

            end;
            c^.Flags := c^.Flags and not $1;
            if not RunBegin(Vars, True) then
              Exit;
          end {if}
          else
          begin
            c^.Flags := c^.Flags or $1;
            for II := IStart to IEnd do begin
              if Assigned(OnRunLine) then
              begin
                RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
                if ErrorCode <> ENoError then Exit;
              end;

              SetInteger(C, II);
              if not RunBegin(Vars, False) then
              begin
                c^.Flags := c^.Flags and not $1;
                Exit;
              end;

              // If a break were returned
              //--------------------------
              if Parser^.CurrTokenId = CSTII_Break
                then lBreak := true;

              Parser^.CurrTokenPos := IPos;
              ParseToken(Parser);

              // If a break were returned
              //--------------------------
              if lBreak then break

            end;
            c^.Flags := c^.Flags and not $1;
            if not RunBegin(Vars, True) then
              Exit;
          end {if}
        end;
//-------------------------------------------------------
// Begin a block
//-------------------------------------------------------
      CSTII_Begin:
        begin
          if not RunBegin(Vars, False) then
            Exit;
        end; {CSTII_Begin}
//-------------------------------------------------------
// CASE <x> OF
//-------------------------------------------------------
      CSTII_Case:
        begin
          SkipWhiteSpace(Parser);
          c := CreateCajVariant(TM_Add(Types, '', CSV_Var, nil));
          c^.CV_Var := nil; {Say that calc can assign any type}
          if not Calc(vars, c, CSTII_Of, False) then
          begin
            DestroyCajVariant(c);
            Exit;
          end; {If}
          if Parser^.CurrTokenId <> CSTII_OF then
          begin
            RunError(EOfExpected);
            exit;
          end; {If}
          SkipWhiteSpace(Parser);
          B := False;
          while Parser^.CurrTokenId <> CSTII_End do begin
            if Parser^.CurrTokenId = CSTII_Else then
            begin
              SkipWhiteSpace(Parser);
              if not RunBegin(Vars, B) then
                Exit;
              if Parser^.CurrTokenId = CSTI_SemiColon then
              begin
                SkipWhiteSpace(Parser);
              end;
              if Parser^.CurrTokenId <> CSTII_End then
              begin
                RunError(EEndExpected);
                Exit;
              end;
              Break;
            end;
            if Assigned(OnRunLine) then
            begin
              RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
              if ErrorCode <> ENoError then
                Exit;
            end;
            c2 := CreateCajVariant(C^.VType);
            if not Calc(vars, c2, CSTI_Colon, False) then
            begin
              DestroyCajVariant(c);
              DestroyCajVariant(c2);
              Exit;
            end; {If}
            if Parser^.CurrTokenId <> CSTI_Colon then
            begin
              RunError(EColonExpected);
              DestroyCajVariant(c);
              DestroyCajVariant(c2);
              Exit;
            end; {If}
            SkipWhiteSpace(Parser);
            if not Perform(Types, c2, c, ptEqual) then
            begin
              RunError(ETypeMismatch);
              DestroyCajVariant(c);
              DestroyCajVariant(c2);
              Exit;
            end; {If}
            if not RunBegin(Vars, (not c2^.CV_Bool or B)) then
            begin
              DestroyCajVariant(c);
              DestroyCajVariant(c2);
              Exit;
            end;
            if c2^.CV_Bool then
              B := True;
            if Parser^.CurrTokenid = CSTI_SemiColon then
            begin
              SkipWhiteSpace(Parser);
            end;
            DestroyCajVariant(c2);
          end; {While}
          DestroyCajVariant(c);
          SkipWhiteSpace(Parser); {Skip end}
        end; {CSTII_Case}

//-------------------------------------------------------
// Found an identifier
//-------------------------------------------------------
      CSTI_Identifier:
        begin
          if Assigned(OnRunLine) then
          begin
            RunError(OnRunLine(fId, self, Parser^.CurrTokenPos));
            if ErrorCode <> ENoError then Exit;
          end;

          if PM_Find(InternalProcedures, FastUppercase(GetToken(Parser))) <> -1
            then
          begin
            DestroyCajVariant(DoProc(Vars, True));
            if ErrorCode <> ENoError then Exit;
          end {if}
          else if Assigned(vars) and (Vm_Find(Vars,
            FastUppercase(GetToken(Parser))) <> -1) then
          begin
            c := GetVarLink(VM_Get(Vars, Vm_Find(Vars,
              FastUppercase(GetToken(Parser)))));
            if (1 and c^.Flags) <> 0 then
            begin
              RunError(EErrorInStatement);
              Exit;
            end; {else if}
            if not CalcSub(Vars, c) then
              Exit;
            SkipWhiteSpace(Parser);
            if Parser^.CurrTokenId <> CSTI_Assignment then
            begin
              RunError(EAssignmentExpected);
              Exit;
            end; {if}
            SkipWhiteSpace(Parser);
            if not Calc(vars, c, CSTI_Semicolon, False) then
              Exit;
          end {if}
          else if Vm_Find(Variables, FastUppercase(GetToken(Parser))) <> -1 then
          begin
            c := GetVarLink(VM_Get(Variables,
              Vm_Find(Variables, FastUppercase(GetToken(Parser)))));
            if (1 and c^.Flags) <> 0 then
            begin
              RunError(EErrorInStatement);
              Exit;
            end; {else if}
            if not CalcSub(Vars, c) then
              Exit;
            SkipWhiteSpace(Parser);
            if Parser^.CurrTokenId <> CSTI_Assignment then
            begin
              RunError(EAssignmentExpected);
              Exit;
            end; {if}
            SkipWhiteSpace(Parser);
            if not Calc(vars, c, CSTI_Semicolon, False) then
              Exit;
          end {if}
          else if PM_Find(Procedures, FastUppercase(GetToken(Parser))) <> -1
            then
          begin
            DestroyCajVariant(DoProc(Vars, False));
            if ErrorCode <> ENoError then
              Exit;
          end {else if}
          else
          begin
            RunError(EUnknownIdentifier);
            Exit;
          end; {if}
        end; {CSTI_Identifier}
    else
      begin
        if Parser^.CurrTokenId <> CSTI_Period
          then RunError(EErrorInStatement);
        Exit;
      end; {Else case}
    end; {Case}
  end; {While}
  RunBegin := True;
end; {RunBegin}

function StdProc(ID: Pointer; const ProcName: string; Params: PVariableManager;
  res: PIfVariant): TCS2Error; {$IFDEF P16}far; {$ENDIF}
var
  c: PIfVariant;
  i1, i2: LongInt;
  cStr: string;
  function mkchr(c: PIfVariant): Integer;
  begin
    if c^.vtype^.atypeid = CSV_String then
    begin
      if Length(c^.cv_str) = 1 then
      begin
        mkChr := Ord(c^.CV_str[1]);
      end else
        mkchr := -1;
    end else
    begin
      mkChr := Ord(c^.CV_Char);
    end;
  end;
begin
  StdProc := ENoError;
  if ProcName = 'STRGET' then
  begin
    c := GetVarLink(VM_Get(params, 0));
    i1 := GetInteger(GetVarLink(VM_Get(params, 1)));
    if (i1 < 1) or (i1 > Length(c^.cv_Str)) then
    begin
      StdProc := ERangeError;
      Exit;
    end;
    Res^.CV_Char := c^.cv_Str[i1];
  end else if ProcName = 'STRSET' then
  begin
    c := GetVarLink(VM_Get(params, 2));
    i1 := GetInteger(GetVarLink(VM_Get(params, 1)));
    if (i1 < 1) or (i1 > Length(c^.cv_Str)) then
    begin
      StdProc := ERangeError;
      Exit;
    end;
    i2 := MkChr(GetVarLink(VM_Get(params, 2)));
    if i2 = -1 then
    begin
      StdProc := ERangeError;
      Exit;
    end;
    C^.CV_Str[i1] := Chr(i2);
  end else if (ProcName = 'ORD') then
  begin
    i1 := MkChr(GetVarLink(VM_Get(params, 0)));
    if i1 = -1 then
    begin
      StdProc := ERangeError;
      Exit;
    end;
    res^.cv_UByte := i1;
  end else if ProcName = 'CHR' then
  begin
    res^.Cv_Char := Chr(GetInteger(GetVarLink(VM_Get(Params, 0))));
  end else if ProcName = 'UPPERCASE' then
  begin
    SetString(Res, FastUppercase(GetString(GetVarLink(VM_Get(Params, 0)))));
  end else if ProcName = 'LOWERCASE' then
  begin
    SetString(Res, FastLowercase(GetString(GetVarLink(VM_Get(Params, 0)))));
  end else if ProcName = 'TRIM' then
  begin
    SetString(Res, FastTrim(GetString(GetVarLink(VM_Get(Params, 0)))));
  end else if (ProcName = 'POS') then
  begin
    SetInteger(Res, Pos(GetString(GetVarLink(VM_Get(Params, 0))),
      GetString(GetVarLink(VM_Get(Params, 1)))));
  end else if ProcName = 'INTTOSTR' then
  begin
    SetString(Res, IntToStr(GetInteger(GetVarLink(VM_Get(Params, 0)))));
  end else if ProcName = 'STRTOINT' then
  begin
    SetInteger(Res, StrToIntDef(GetString(GetVarLink(VM_Get(Params, 0))),
      GetInteger(GetVarLink(VM_Get(Params, 1)))));
  end else if ProcName = 'COPY' then
  begin
    SetString(Res, Copy(GetString(GetVarLink(VM_Get(Params, 0))),
      GetInteger(GetVarLink(VM_Get(Params, 1))),
      GetInteger(GetVarLink(VM_Get(Params, 2)))));
  end else if ProcName = 'LEFT' then
  begin
    SetString(Res, Copy(GetString(GetVarLink(VM_Get(Params, 0))), 1,
      GetInteger(GetVarLink(VM_Get(Params, 1)))));
  end else if ProcName = 'DELETE' then
  begin
    c := GetVarLink(VM_Get(params, 0));
    Delete(c^.cv_Str, GetInteger(GetVarLink(VM_Get(Params, 1))),
      GetInteger(GetVarLink(VM_Get(Params, 2))));
  end else if ProcName = 'INSERT' then
  begin
    c := GetVarLink(VM_Get(params, 1));
    Insert(GetString(GetVarLink(VM_Get(Params, 0))), c^.cv_Str,
      GetInteger(GetVarLink(VM_Get(Params, 2))));
  end else if ProcName = 'SETARRAYLENGTH' then
  begin
    c := GetVarLink(VM_Get(params, 0));
    i1 := GetInteger(GetVarLink(VM_Get(params, 1)));
    if i1 > c^.CV_ArrItems.Count then
    begin
      for i2 := 1 to i1 - C^.CV_ArrItems.Count do begin
        c^.CV_ArrItems.Add(CreateCajVariant(c^.VType^.ext));
      end;
    end else if i1 < c^.CV_ArrItems.Count then
    begin
      for i2 := 1 to c^.CV_ArrItems.Count - i1 do begin
        DestroyCajVariant(c^.CV_ArrItems.GetItem(c^.Cv_ArrItems.Count - 1));
        c^.Cv_ArrItems.Delete(c^.Cv_ArrItems.Count - 1);
      end;
    end;
  end else if ProcName = 'GETARRAYLENGTH' then
  begin
    SetInteger(Res, GetVarLink(VM_Get(params, 0))^.CV_ArrItems.Count);
  end else if ProcName = 'LENGTH' then
  begin
    SetInteger(Res, Length(GetString(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'SIN' then
  begin
    SetReal(Res, Sin(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'COS' then
  begin
    SetReal(Res, Cos(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'SQRT' then
  begin
    SetReal(Res, Sqrt(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'PI' then
  begin
    SetReal(Res, PI);
  end else if ProcName = 'ROUND' then
  begin
    SetInteger(Res, Round(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'TRUNC' then
  begin
    SetInteger(Res, Trunc(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'INT' then
  begin
    SetInteger(Res, Trunc(GetReal(GetVarLink(VM_Get(params, 0))) + 0.5));
  end else if ProcName = 'ABS' then
  begin
    SetReal(Res, Abs(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'SQRT' then
  begin
    SetReal(Res, Sqrt(GetReal(GetVarLink(VM_Get(params, 0)))));
  end else if ProcName = 'FLOATTOSTR' then
  begin
    SetString(Res, FloatToStr(GetReal(GetVarLink(VM_Get(Params, 0)))));
  end else if ProcName = 'PADZ' then
  begin
    SetString(Res, Padz(GetString(GetVarLink(VM_Get(Params, 0))),
      GetInteger(GetVarLink(VM_Get(Params, 1)))));
  end else if ProcName = 'PADL' then
  begin
    SetString(Res, Padl(GetString(GetVarLink(VM_Get(Params, 0))),
      GetInteger(GetVarLink(VM_Get(Params, 1)))));
  end else if ProcName = 'PADR' then
  begin
    SetString(Res, Padr(GetString(GetVarLink(VM_Get(Params, 0))),
      GetInteger(GetVarLink(VM_Get(Params, 1)))));
  end else if ProcName = 'VERSION' then
  begin
    SetString(Res, Version);
  end else if (ProcName = 'REPLICATE') or (ProcName = 'STRINGOFCHAR') then
  begin
    cStr := GetString(GetVarLink(VM_Get(Params, 0)));
    SetString(Res, StringOfChar(cStr[1],
      GetInteger(GetVarLink(VM_Get(Params, 1)))));

  end;

end;

//
//  Purpose: Adds a pointer to a needed resource
//
//-------------------------------------------------------------------

procedure TIfPasScript.AddResource(FreeProc: TResourceFree; Data: Pointer);
var
  p: PResourceData;
begin
  new(p);
  p^.data := data;
  p^.freeProc := FreeProc;
  FAllocatedResources.Add(p);
end;

//
//  Purpose: Removes a resource pointer and cleans up the memory
//
//-------------------------------------------------------------------

procedure TIfPasScript.RemoveResource(Data: Pointer);
var
  i: Longint;
  p: PResourceData;
begin
  for i := FAllocatedResources.Count - 1 downto 0 do
  begin
    p := FAllocatedResources.GetItem(i);
    if p^.data = data then
    begin
      FAllocatedResources.Delete(i);
      Dispose(p);
      break;
    end;
  end;
end;

//-------------------------------------------------------------------

function TIfPasScript.IsValidResource(FreeProc: TResourceFree; Data: Pointer): Boolean;
var
  i: Longint;
  p: PResourceData;
begin
  result := true;

  for i := FAllocatedResources.Count - 1 downto 0 do
  begin
    p := FAllocatedResources.GetItem(i);
    if (p^.data = data) and (@p^.FreeProc = @FreeProc)
      then exit;
  end;
  result := false;
end;

function TIfPasScript.AddFunction(Proc: Pointer; Decl: String): Boolean;
var
  Parser : PIfPascalParser;
  CurrVar : String;
  FuncName,
  FuncParam : String;
  FuncRes,
  CurrType : Longint;
  function GetType(const s: string): Longint;
  begin
    Result := Longint(TM_Get(Types, s));
  end;
Begin
  Decl:=Decl+#0;
  New(Parser);
  Parser^.CurrTokenPos:=0;
  Parser^.text:=@Decl[1];
  ParseToken(Parser);
  If Parser^. CurrTokenId = CSTII_Procedure Then
    FuncRes := 0
  Else
    FuncRes := 1;
  SkipWhiteSpace(Parser);
  FuncName := FastUppercase (GetToken (Parser) );
  FuncParam := FuncName;
  CurrVar := '';
  SkipWhiteSpace(Parser);
  If parser^. CurrTokenId = CSTI_OpenRound Then Begin
    SkipWhiteSpace(Parser);
    While True Do Begin
      If Parser^. CurrTokenId = CSTII_Var Then Begin
        CurrVar := '!';
        SkipWhiteSpace(Parser);
      End; {if}
      While True Do Begin
        CurrVar := CurrVar + fastuppercase (GetToken (Parser) ) + '|';
        SkipWhiteSpace(parser);
        If Parser^. CurrTokenId = CSTI_Colon Then Break;
        SkipWhiteSpace(Parser);
      End; {while}
      SkipWhiteSpace(Parser);
      CurrType := GetType (FastUppercase (GetToken (Parser) ) );
      If Pos ('!', CurrVar) = 1 Then Begin
        Delete (currVar, 1, 1);
        While Pos ('|', CurrVar) > 0 Do Begin
          FuncParam := FuncParam + ' !' + Copy (CurrVar, 1, Pos ('|', CurrVar) - 1) + ' ' + IntToStr (CurrType);
          Delete (CurrVar, 1, Pos ('|', CurrVar) );
        End; {while}
      End Else Begin
        While Pos ('|', CurrVar) > 0 Do Begin
          FuncParam := FuncParam + ' ' + Copy (CurrVar, 1, Pos ('|', CurrVar) - 1) + ' ' + IntToStr (CurrType);
          Delete (CurrVar, 1, Pos ('|', CurrVar) );
        End; {while}
      End; {if}
      SkipWhiteSpace(Parser);
      If Parser^. CurrTokenId = CSTI_CloseRound Then Begin
        SkipWhiteSpace(Parser);
        Break;
      End; {if}
      SkipWhiteSpace(Parser);
    End;
  End;
  If FuncRes = 1 Then Begin
    SkipWhiteSpace(Parser);
    FuncRes :=  GetType (FastUppercase (GetToken (Parser) ) );
  End;
  FuncParam := InttoStr (FuncRes) + ' ' + FuncParam;
  PM_Add(Procedures, FuncParam, Proc);
  Dispose(Parser);
  Result := True;
end;


//
//  Purpose: Adds various internal function calls
//
//-------------------------------------------------------------------

procedure RegisterStdLib(P: TIfPasScript);
{Register standard library}
begin
  P.AddFunction(@StdProc, 'function StrGet(var s: string; I: Longint): char');
  P.AddFunction(@StdProc, 'procedure StrSet(c: char; i: Longint; var s: string): char');
  P.AddFunction(@StdProc, 'function Ord(c: char): Byte');
  P.AddFunction(@StdProc, 'function Chr(b: Byte): Char');
  P.AddFunction(@StdProc, 'function StrToInt(s: string; I: Longint): Integer');
  P.AddFunction(@StdProc, 'function IntToStr(i: Longint): string');
  P.AddFunction(@StdProc, 'function Uppercase(s: string): string');
  P.AddFunction(@StdProc, 'function Copy(s: string; i1, i2: Longint): string');
  P.AddFunction(@StdProc, 'procedure Delete(var s: string; i1,i2: Longint)');
  P.AddFunction(@StdProc, 'procedure Insert(s1: string; var s: string; i1: Longint)');
  P.AddFunction(@StdProc, 'function Pos(s1, s2: string): Longint');
  P.AddFunction(@StdProc, 'function Length(s: string): Longint');
  P.AddFunction(@StdProc, 'function sin(s: Extended): Extended');
  P.AddFunction(@StdProc, 'function cos(s: Extended): Extended');
  P.AddFunction(@StdProc, 'function SQRT(s: Extended): Extended');
  P.AddFunction(@StdProc, 'function Round(s: Extended): Longint');
  P.AddFunction(@StdProc, 'function Trunc(s: Extended): Longint');
  P.AddFunction(@StdProc, 'function PI: Extended');
  P.AddFunction(@StdProc, 'function LowerCase(s: string): string');
  P.AddFunction(@StdProc, 'function Trim(s: string): string');
  P.AddFunction(@StdProc, 'function Int(s: Extended): Longint');
  P.AddFunction(@StdProc, 'function FloatToStr(s: Extended): string');
  P.AddFunction(@StdProc, 'function replicate(s: string; i: Longint): string');
  P.AddFunction(@StdProc, 'function Version: string');
  P.AddFunction(@StdProc, 'function Left(s: string; i: Longint): string');
  P.AddFunction(@StdProc, 'function StringOfChar(s: string; i: Integer): string');
  p.AddFunction(@StdProc, 'procedure SetArrayLength(var u: array; Length: Longint)');
  p.AddFunction(@StdProc, 'function GetArrayLength(var u: array): Longint');
end;
//-------------------------------------------------------------------
end.

